"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.team_pb2 import (
    CreateEnvironmentRequest,
    CreateEnvironmentResponse,
    CreateProjectRequest,
    CreateProjectResponse,
    CreateServiceTokenRequest,
    CreateServiceTokenResponse,
    CreateTeamRequest,
    CreateTeamResponse,
    DeleteServiceTokenRequest,
    DeleteServiceTokenResponse,
    ExpireTeamInviteRequest,
    ExpireTeamInviteResponse,
    GetAgentRequest,
    GetAgentResponse,
    GetAvailablePermissionsRequest,
    GetAvailablePermissionsResponse,
    GetDisplayAgentRequest,
    GetDisplayAgentResponse,
    GetEnvRequest,
    GetEnvResponse,
    GetEnvironmentsRequest,
    GetEnvironmentsResponse,
    GetTeamPermissionsRequest,
    GetTeamPermissionsResponse,
    GetTeamRequest,
    GetTeamResponse,
    InviteTeamMemberRequest,
    InviteTeamMemberResponse,
    ListServiceTokensRequest,
    ListServiceTokensResponse,
    ListTeamInvitesRequest,
    ListTeamInvitesResponse,
    UpdateEnvironmentRequest,
    UpdateEnvironmentResponse,
    UpdateScimGroupSettingsRequest,
    UpdateScimGroupSettingsResponse,
    UpdateServiceTokenRequest,
    UpdateServiceTokenResponse,
    UpsertFeaturePermissionsRequest,
    UpsertFeaturePermissionsResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class TeamServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetEnv: UnaryUnaryMultiCallable[
        GetEnvRequest,
        GetEnvResponse,
    ]
    GetEnvironments: UnaryUnaryMultiCallable[
        GetEnvironmentsRequest,
        GetEnvironmentsResponse,
    ]
    GetAgent: UnaryUnaryMultiCallable[
        GetAgentRequest,
        GetAgentResponse,
    ]
    GetDisplayAgent: UnaryUnaryMultiCallable[
        GetDisplayAgentRequest,
        GetDisplayAgentResponse,
    ]
    GetTeam: UnaryUnaryMultiCallable[
        GetTeamRequest,
        GetTeamResponse,
    ]
    CreateTeam: UnaryUnaryMultiCallable[
        CreateTeamRequest,
        CreateTeamResponse,
    ]
    CreateProject: UnaryUnaryMultiCallable[
        CreateProjectRequest,
        CreateProjectResponse,
    ]
    CreateEnvironment: UnaryUnaryMultiCallable[
        CreateEnvironmentRequest,
        CreateEnvironmentResponse,
    ]
    UpdateEnvironment: UnaryUnaryMultiCallable[
        UpdateEnvironmentRequest,
        UpdateEnvironmentResponse,
    ]
    GetAvailablePermissions: UnaryUnaryMultiCallable[
        GetAvailablePermissionsRequest,
        GetAvailablePermissionsResponse,
    ]
    CreateServiceToken: UnaryUnaryMultiCallable[
        CreateServiceTokenRequest,
        CreateServiceTokenResponse,
    ]
    DeleteServiceToken: UnaryUnaryMultiCallable[
        DeleteServiceTokenRequest,
        DeleteServiceTokenResponse,
    ]
    ListServiceTokens: UnaryUnaryMultiCallable[
        ListServiceTokensRequest,
        ListServiceTokensResponse,
    ]
    UpdateServiceToken: UnaryUnaryMultiCallable[
        UpdateServiceTokenRequest,
        UpdateServiceTokenResponse,
    ]
    InviteTeamMember: UnaryUnaryMultiCallable[
        InviteTeamMemberRequest,
        InviteTeamMemberResponse,
    ]
    ExpireTeamInvite: UnaryUnaryMultiCallable[
        ExpireTeamInviteRequest,
        ExpireTeamInviteResponse,
    ]
    ListTeamInvites: UnaryUnaryMultiCallable[
        ListTeamInvitesRequest,
        ListTeamInvitesResponse,
    ]
    UpsertFeaturePermissions: UnaryUnaryMultiCallable[
        UpsertFeaturePermissionsRequest,
        UpsertFeaturePermissionsResponse,
    ]
    UpdateScimGroupSettings: UnaryUnaryMultiCallable[
        UpdateScimGroupSettingsRequest,
        UpdateScimGroupSettingsResponse,
    ]
    GetTeamPermissions: UnaryUnaryMultiCallable[
        GetTeamPermissionsRequest,
        GetTeamPermissionsResponse,
    ]
    """ rpc SetEnvironmentVariables(SetEnvironmentVariablesRequest) returns
     (SetEnvironmentVariablesResponse) {
       option (chalk.auth.v1.permission) = PERMISSION_DEPLOY_CREATE;
     }
    """

class TeamServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetEnv(
        self,
        request: GetEnvRequest,
        context: ServicerContext,
    ) -> GetEnvResponse: ...
    @abstractmethod
    def GetEnvironments(
        self,
        request: GetEnvironmentsRequest,
        context: ServicerContext,
    ) -> GetEnvironmentsResponse: ...
    @abstractmethod
    def GetAgent(
        self,
        request: GetAgentRequest,
        context: ServicerContext,
    ) -> GetAgentResponse: ...
    @abstractmethod
    def GetDisplayAgent(
        self,
        request: GetDisplayAgentRequest,
        context: ServicerContext,
    ) -> GetDisplayAgentResponse: ...
    @abstractmethod
    def GetTeam(
        self,
        request: GetTeamRequest,
        context: ServicerContext,
    ) -> GetTeamResponse: ...
    @abstractmethod
    def CreateTeam(
        self,
        request: CreateTeamRequest,
        context: ServicerContext,
    ) -> CreateTeamResponse: ...
    @abstractmethod
    def CreateProject(
        self,
        request: CreateProjectRequest,
        context: ServicerContext,
    ) -> CreateProjectResponse: ...
    @abstractmethod
    def CreateEnvironment(
        self,
        request: CreateEnvironmentRequest,
        context: ServicerContext,
    ) -> CreateEnvironmentResponse: ...
    @abstractmethod
    def UpdateEnvironment(
        self,
        request: UpdateEnvironmentRequest,
        context: ServicerContext,
    ) -> UpdateEnvironmentResponse: ...
    @abstractmethod
    def GetAvailablePermissions(
        self,
        request: GetAvailablePermissionsRequest,
        context: ServicerContext,
    ) -> GetAvailablePermissionsResponse: ...
    @abstractmethod
    def CreateServiceToken(
        self,
        request: CreateServiceTokenRequest,
        context: ServicerContext,
    ) -> CreateServiceTokenResponse: ...
    @abstractmethod
    def DeleteServiceToken(
        self,
        request: DeleteServiceTokenRequest,
        context: ServicerContext,
    ) -> DeleteServiceTokenResponse: ...
    @abstractmethod
    def ListServiceTokens(
        self,
        request: ListServiceTokensRequest,
        context: ServicerContext,
    ) -> ListServiceTokensResponse: ...
    @abstractmethod
    def UpdateServiceToken(
        self,
        request: UpdateServiceTokenRequest,
        context: ServicerContext,
    ) -> UpdateServiceTokenResponse: ...
    @abstractmethod
    def InviteTeamMember(
        self,
        request: InviteTeamMemberRequest,
        context: ServicerContext,
    ) -> InviteTeamMemberResponse: ...
    @abstractmethod
    def ExpireTeamInvite(
        self,
        request: ExpireTeamInviteRequest,
        context: ServicerContext,
    ) -> ExpireTeamInviteResponse: ...
    @abstractmethod
    def ListTeamInvites(
        self,
        request: ListTeamInvitesRequest,
        context: ServicerContext,
    ) -> ListTeamInvitesResponse: ...
    @abstractmethod
    def UpsertFeaturePermissions(
        self,
        request: UpsertFeaturePermissionsRequest,
        context: ServicerContext,
    ) -> UpsertFeaturePermissionsResponse: ...
    @abstractmethod
    def UpdateScimGroupSettings(
        self,
        request: UpdateScimGroupSettingsRequest,
        context: ServicerContext,
    ) -> UpdateScimGroupSettingsResponse: ...
    @abstractmethod
    def GetTeamPermissions(
        self,
        request: GetTeamPermissionsRequest,
        context: ServicerContext,
    ) -> GetTeamPermissionsResponse:
        """rpc SetEnvironmentVariables(SetEnvironmentVariablesRequest) returns
        (SetEnvironmentVariablesResponse) {
          option (chalk.auth.v1.permission) = PERMISSION_DEPLOY_CREATE;
        }
        """

def add_TeamServiceServicer_to_server(servicer: TeamServiceServicer, server: Server) -> None: ...



from __future__ import annotations

import collections
import dataclasses
import json
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, List, Optional, Type, TypeVar, Union, cast, overload

from chalk import Resolver, Windowed
from chalk.features import DataFrame, Feature, Features, Filter, serialize_dtype
from chalk.features.feature_field import HasOnePathObj, VersionInfo
from chalk.features.feature_set import is_feature_set_class
from chalk.features.pseudofeatures import PSEUDONAMESPACE
from chalk.features.resolver import Cron, OfflineResolver, OnlineResolver, SinkResolver, StreamResolver
from chalk.parsed.duplicate_input_gql import (
    FeatureClassGQL,
    RecomputeFeaturesGQL,
    UpsertCDCSourceGQL,
    UpsertCronQueryGQL,
    UpsertDataFrameGQL,
    UpsertFeatureGQL,
    UpsertFeatureIdGQL,
    UpsertFeatureReferenceGQL,
    UpsertFeatureTimeKindGQL,
    UpsertFilterGQL,
    UpsertHasManyKindGQL,
    UpsertHasOneKindGQL,
    UpsertNamedQueryGQL,
    UpsertReferencePathComponentGQL,
    UpsertResolverGQL,
    UpsertResolverInputUnionGQL,
    UpsertResolverOutputGQL,
    UpsertScalarKindGQL,
    UpsertSinkResolverGQL,
    UpsertStreamResolverGQL,
    UpsertStreamResolverParamGQL,
    UpsertStreamResolverParamKeyedStateGQL,
    UpsertStreamResolverParamMessageGQL,
    UpsertWindowMaterializationGQL,
    VersionInfoGQL,
)
from chalk.queries.named_query import NamedQuery
from chalk.queries.scheduled_query import ScheduledQuery
from chalk.sql._internal.sql_source import BaseSQLSource, TableIngestMixIn
from chalk.streams.types import (
    StreamResolverParam,
    StreamResolverParamKeyedState,
    StreamResolverParamMessage,
    StreamResolverParamMessageWindow,
)
from chalk.utils import paths
from chalk.utils.collections import get_unique_item, unwrap_optional_and_annotated_if_needed
from chalk.utils.duration import CronTab, Duration, timedelta_to_duration
from chalk.utils.json import JSON
from chalk.utils.pydanticutil.pydantic_compat import (
    get_pydantic_model_json,
    get_pydantic_output_structure,
    is_pydantic_basemodel,
    is_pydantic_basemodel_instance,
)
from chalk.utils.string import to_snake_case

if TYPE_CHECKING:
    import pydantic
else:
    try:
        import pydantic.v1 as pydantic
    except ImportError:
        import pydantic

T = TypeVar("T")


@dataclasses.dataclass
class _ConvertedType:
    name: str
    bases: List[str]


def _get_qualified_class_name(cls: Type[Any]):
    mod = cls.__module__
    return cls.__qualname__ if mod == "builtins" else f"{mod}.{cls.__qualname__}"


def _convert_type(underlying_type: Type, is_windowed: bool) -> _ConvertedType:
    if underlying_type is JSON:  # pyright: ignore[reportUnnecessaryComparison]
        return _ConvertedType(
            name="chalk.JSON",
            bases=["object"],
        )
    if isinstance(underlying_type, Windowed):
        is_windowed = True
        underlying_type = underlying_type.kind

    pretty_name = repr(underlying_type)
    if isinstance(underlying_type, type):  # pyright: ignore[reportUnnecessaryIsInstance]
        pretty_name = underlying_type.__name__

    bases = []
    actual_underlying_type = unwrap_optional_and_annotated_if_needed(underlying_type)
    if isinstance(actual_underlying_type, type):  # pyright: ignore[reportUnnecessaryIsInstance]
        bases.extend(x.__name__ for x in actual_underlying_type.mro())
    else:
        bases = ["object"]

    pretty_name = pretty_name.replace("typing.", "")
    if is_windowed:
        bases.append("CHALK_WINDOWED")
        return _ConvertedType(name=f"Windowed[{pretty_name}]", bases=bases)
    if isinstance(underlying_type, Feature):
        return _convert_type(underlying_type.converter.rich_type, underlying_type.is_windowed)

    return _ConvertedType(
        name=pretty_name,
        bases=bases,
    )


def _get_feature_id(s: Feature, root_fqn: bool):
    assert s.features_cls is not None
    return UpsertFeatureIdGQL(
        fqn=s.root_fqn if root_fqn else s.fqn,
        name=s.name,
        namespace=s.namespace,
        isPrimary=s.primary,
        className=s.features_cls.__name__,
        attributeName=s.attribute_name,
        explicitNamespace=s.namespace != to_snake_case(s.features_cls.__name__),
    )


def _get_pseudofeature_id(s: Feature):
    return UpsertFeatureIdGQL(
        fqn=s.fqn,
        name=s.name,
        namespace=s.namespace,
        isPrimary=s.primary,
    )


def _convert_df(df: Type[DataFrame]) -> UpsertDataFrameGQL:
    return UpsertDataFrameGQL(
        columns=[
            _get_feature_id(f, root_fqn=True)
            for f in df.columns
            if not f.is_autogenerated and not f.namespace == PSEUDONAMESPACE
        ],
        # filters=[convert_type_to_gql(f) for f in df.filters],
        filters=None,
    )


def _get_path_component(pc: HasOnePathObj) -> UpsertReferencePathComponentGQL:
    assert isinstance(pc.parent, Feature), f"Parent in relationship path not a feature, but {type(pc).__name__}"
    return UpsertReferencePathComponentGQL(
        parent=_get_feature_id(pc.parent, root_fqn=False),
        child=_get_feature_id(pc.child, root_fqn=False),
        parentToChildAttributeName=pc.parent_to_child_attribute_name,
    )


def _parse_generic_params_recursively(typ: Any) -> str:
    """
    Stringify a type with generic parameters like 'list[list[str]]',
    making sure parameters are included recursively.
    :return: the type name along with any parameters it may have
    """
    if not isinstance(typ, type):
        type_name = repr(typ)
    else:
        type_name = typ.__name__

    type_name = type_name.removeprefix("typing.")
    if not hasattr(typ, "__args__"):
        return type_name
    params = (_parse_generic_params_recursively(x) for x in typ.__args__)
    param_string = ",".join(params)
    return f"{type_name}[{param_string}]"


@overload
def convert_type_to_gql(t: Filter, path_prefix: Optional[str] = None) -> UpsertFilterGQL:
    ...


@overload
def convert_type_to_gql(t: StreamResolver, path_prefix: Optional[str] = None) -> UpsertStreamResolverGQL:
    ...


@overload
def convert_type_to_gql(t: SinkResolver, path_prefix: Optional[str] = None) -> UpsertSinkResolverGQL:
    ...


@overload
def convert_type_to_gql(
    t: Union[OnlineResolver, OfflineResolver], path_prefix: Optional[str] = None
) -> UpsertResolverGQL:
    ...


@overload
def convert_type_to_gql(
    t: Resolver, path_prefix: Optional[str] = None
) -> UpsertStreamResolverGQL | UpsertSinkResolverGQL | UpsertResolverGQL:
    ...


@overload
def convert_type_to_gql(t: StreamResolverParam, path_prefix: Optional[str] = None) -> UpsertStreamResolverParamGQL:
    ...


@overload
def convert_type_to_gql(t: ScheduledQuery, path_prefix: Optional[str] = None) -> UpsertCronQueryGQL:
    ...


@overload
def convert_type_to_gql(t: NamedQuery, path_prefix: Optional[str] = None) -> UpsertNamedQueryGQL:
    ...


@overload
def convert_type_to_gql(t: Feature, path_prefix: Optional[str] = None) -> UpsertFeatureGQL:
    ...


@overload
def convert_type_to_gql(t: Type[Features], path_prefix: Optional[str] = None) -> FeatureClassGQL:
    ...


def convert_type_to_gql(
    t: Any,
    # Resolvers have their filenames attached to them. But for UI presentation, we don't
    # want to show the full path, e.g. /Users/customer-name/projects/chalk/....,
    # but we want to show just the path relative to the project root. To do this efficiently,
    # we'll pass the project root path here and delete it from those full paths
    path_prefix: Optional[str] = None,
):
    if isinstance(t, StreamResolver):
        return UpsertStreamResolverGQL(
            fqn=t.fqn,
            kind="stream",
            sourceClassName=_get_qualified_class_name(t.source.__class__),
            sourceConfig=t.source.config_to_json(),
            functionDefinition=t.function_definition or "",
            environment=None if t.environment is None else list(t.environment),
            doc=t.fn.__doc__,
            machineType=t.machine_type,
            output=[
                _get_feature_id(f, root_fqn=True)
                for f in t.output_features
                if (not isinstance(f, type) or not issubclass(f, DataFrame)) and not f.is_autogenerated
            ],
            inputs=[
                convert_type_to_gql(i)
                for i in t.signature.params
                if not isinstance(i, StreamResolverParamMessageWindow)
            ],
            owner=t.owner,
            filename=t.filename if path_prefix is None else t.filename.replace(path_prefix, ""),
            sourceLine=t.source_line,
        )

    if isinstance(t, SinkResolver):
        return UpsertSinkResolverGQL(
            fqn=t.fqn,
            functionDefinition=t.function_definition or "",
            inputs=[
                UpsertFeatureReferenceGQL(
                    underlying=_get_feature_id(f, root_fqn=False),
                    path=[_get_path_component(p) for p in f.path or []],
                )
                for f in t.inputs
                if isinstance(f, Feature) and not f.is_autogenerated
            ],
            environment=None if t.environment is None else list(t.environment),
            tags=None if t.tags is None else list(t.tags),
            doc=t.doc,
            machineType=t.machine_type,
            bufferSize=t.buffer_size,
            debounce=None if t.debounce is None else timedelta_to_duration(t.debounce),
            maxDelay=None if t.max_delay is None else timedelta_to_duration(t.max_delay),
            upsert=t.upsert,
            owner=t.owner,
            filename=t.filename if path_prefix is None else t.filename.replace(path_prefix, ""),
            sourceLine=t.source_line,
        )

    if isinstance(t, (OnlineResolver, OfflineResolver)):
        cron = t.cron
        if isinstance(cron, Cron):
            assert cron.schedule is not None, "`Cron` object must be constructed with a `schedule` property."
            cron = cron.schedule
        inputs = []
        all_inputs = []
        for f in t.inputs:
            if (
                not isinstance(f, type)
                or not issubclass(f, DataFrame)  # pyright: ignore[reportUnnecessaryIsInstance]  # noqa
            ) and not f.is_autogenerated:  # pyright: ignore[reportAttributeAccessIssue]
                f = cast(Feature, f)
                if f.namespace == PSEUDONAMESPACE:
                    feature_ref = UpsertFeatureReferenceGQL(
                        underlying=_get_pseudofeature_id(f),
                        path=[_get_path_component(p) for p in f.path or []],
                    )
                    all_inputs.append(UpsertResolverInputUnionGQL(pseudoFeature=feature_ref))
                else:
                    feature_ref = UpsertFeatureReferenceGQL(
                        underlying=_get_feature_id(f, root_fqn=False),
                        path=[_get_path_component(p) for p in f.path or []],
                    )
                    inputs.append(feature_ref)
                    all_inputs.append(UpsertResolverInputUnionGQL(feature=feature_ref))
            elif isinstance(f, type) and issubclass(f, DataFrame):  # pyright: ignore[reportUnnecessaryIsInstance]
                all_inputs.append(UpsertResolverInputUnionGQL(dataframe=_convert_df(f)))

        assert t.output is not None
        if t.data_sources is not None:
            data_sources = [source.to_json() for source in t.data_sources]
        else:
            data_sources = None
        return UpsertResolverGQL(
            fqn=t.fqn,
            kind="offline" if isinstance(t, OfflineResolver) else "online",
            functionDefinition=t.function_definition or "",
            inputs=inputs,
            allInputs=all_inputs,
            output=UpsertResolverOutputGQL(
                features=[
                    _get_feature_id(f, root_fqn=True)
                    for f in t.output.features
                    if (not isinstance(f, type) or not issubclass(f, DataFrame)) and not f.is_autogenerated
                ],
                dataframes=[
                    _convert_df(cast("type[DataFrame]", f))
                    for f in t.output.features
                    if isinstance(f, type) and issubclass(f, DataFrame)
                ],
            ),
            environment=None if t.environment is None else list(t.environment),
            tags=None if t.tags is None else list(t.tags),
            doc=t.doc,
            cron=convert_type_crontab(cron) if cron is not None else cron,
            machineType=t.machine_type,
            owner=t.owner,
            timeout=None if t.timeout is None else timedelta_to_duration(t.timeout),
            filename=t.filename if path_prefix is None else t.filename.replace(path_prefix, ""),
            sourceLine=t.source_line,
            dataSources=data_sources,
            dataLineage=t.data_lineage,
        )

    if isinstance(t, StreamResolverParam):
        if isinstance(t, StreamResolverParamMessage):
            converted_type = _convert_type(t.typ, is_windowed=False)
            schema = None
            if is_pydantic_basemodel(t.typ):
                schema = json.loads(get_pydantic_output_structure(t.typ))
            return UpsertStreamResolverParamGQL(
                message=UpsertStreamResolverParamMessageGQL(
                    name=t.name,
                    typeName=converted_type.name,
                    bases=converted_type.bases,
                    schema=schema,
                ),
                state=None,
            )
        elif isinstance(t, StreamResolverParamKeyedState):
            converted_type = _convert_type(t.typ, is_windowed=False)
            default_value = None
            schema = None
            if dataclasses.is_dataclass(t.default_value):
                default_value = json.loads(json.dumps(dataclasses.asdict(cast(Any, t.default_value))))
                schema = pydantic.dataclasses.dataclass(t.typ).__pydantic_model__.schema()
            elif is_pydantic_basemodel_instance(t.default_value):
                assert is_pydantic_basemodel(t.typ)
                default_value = json.loads(get_pydantic_model_json(t.default_value))
                schema = get_pydantic_output_structure(t.typ)
            return UpsertStreamResolverParamGQL(
                state=UpsertStreamResolverParamKeyedStateGQL(
                    name=t.name,
                    defaultValue=default_value,
                    typeName=converted_type.name,
                    bases=converted_type.bases,
                    schema=schema,
                ),
                message=None,
            )

    if isinstance(t, ScheduledQuery):
        if t.errors:
            error_string = "\n".join(f" - {e}" for e in t.errors)
            raise ValueError(f"Found errors in ScheduledQuery '{t.name}':\n{error_string}")
        get_unique_item((f.split(".")[0] for f in t.output), "output")

        return UpsertCronQueryGQL(
            name=t.name,
            cron=convert_type_crontab(t.cron),
            filename=t.filename if path_prefix is None else t.filename.replace(path_prefix, ""),
            output=list(t.output),
            maxSamples=t.max_samples,
            recomputeFeatures=RecomputeFeaturesGQL(
                featureFqns=list(t.recompute_features) if not isinstance(t.recompute_features, bool) else None,
                all=t.recompute_features if isinstance(t.recompute_features, bool) else None,
            ),
            lowerBound=None,
            upperBound=None,
            lowerBoundStr=datetime.isoformat(t.lower_bound) if t.lower_bound is not None else None,
            upperBoundStr=datetime.isoformat(t.upper_bound) if t.upper_bound is not None else None,
            tags=list(t.tags) if t.tags is not None else None,
            requiredResolverTags=list(t.required_resolver_tags) if t.required_resolver_tags is not None else None,
            storeOnline=t.store_online,
            storeOffline=t.store_offline,
            incrementalSources=None if t.incremental_resolvers is None else list(t.incremental_resolvers),
            resourceGroup=t.resource_group,
            plannerOptions=t.planner_options,
        )

    if isinstance(t, NamedQuery):
        if t.errors:
            print_name = t.name + "" if t.version is None else f" ({t.version})"
            error_string = "\n".join(f" - {e}" for e in t.errors)
            raise ValueError(f"Found errors in NamedQuery '{print_name}':\n{t.code}\n{error_string}")
        if t.planner_options is not None:
            try:
                parsed_planner_options = {k: str(v) for k, v in t.planner_options.items()}
            except Exception as e:
                raise ValueError(f"Could not parse provided planner options '{t.planner_options}'") from e
        else:
            parsed_planner_options = None
        if t.filename is None:
            print_name = t.name + "" if t.version is None else f" ({t.version})"
            raise ValueError(f"Found errors in NamedQuery '{print_name}':\n failed to find filename")

        return UpsertNamedQueryGQL(
            name=t.name,
            queryVersion=t.version,
            input=t.input,
            output=t.output,
            additionalLoggedFeatures=t.additional_logged_features,
            filename=t.filename if path_prefix is None else t.filename.replace(path_prefix, ""),
            description=t.description,
            owner=t.owner,
            meta=None if t.meta is None else dict(t.meta),
            tags=list(t.tags) if t.tags is not None else None,
            staleness=t.staleness,
            plannerOptions=parsed_planner_options,
            code=t.code,
            sourceLineStart=t.source_line_end,
            sourceLineEnd=t.source_line_start,
            validPlanNotRequired=t.valid_plan_not_required,
        )

    if isinstance(t, Feature):
        assert t.name is not None, "Feature has no name"
        assert t.namespace is not None, "Feature has no namespace"
        assert t.features_cls is not None
        scalar_kind_gql = None
        has_one_kind_gql = None
        has_many_kind_gql = None
        feature_time_kind_gql = None
        if t.is_autogenerated:
            raise RuntimeError("Autogenerated features should not be converted")

        if t.is_has_one:
            if t.join is None:
                builder = t.features_cls.__chalk_error_builder__
                builder.add_diagnostic(
                    range=builder.annotation_range(t.attribute_name),
                    message=(
                        f"The attribute '{t.features_cls.__name__}.{t.attribute_name}' "
                        f"references {t.typ}. However, the has-one relationship is not defined."
                    ),
                    raise_error=TypeError,
                    code_href="https://docs.chalk.ai/docs/has-one",
                    label="Missing has-one join",
                    code="35",
                )
            assert t.join is not None
            has_one_kind_gql = UpsertHasOneKindGQL(join=convert_type_to_gql(t.join))

        elif t.is_has_many:
            if t.join is None:
                builder = t.features_cls.__chalk_error_builder__
                builder.add_diagnostic(
                    range=builder.annotation_range(t.attribute_name),
                    message=(
                        f"The attribute '{t.features_cls.__name__}.{t.attribute_name}' "
                        f"references a DataFrame. However, the has-many relationship is not defined."
                    ),
                    label="Missing has-many join",
                    raise_error=TypeError,
                    code="34",
                    code_href="https://docs.chalk.ai/docs/has-many",
                )
            has_many_kind_gql = UpsertHasManyKindGQL(
                join=convert_type_to_gql(t.join),
                columns=None,
                filters=None,
            )

        elif t.join is not None:
            # If a has_one/has_many has an incorrect type annotation
            builder = t.features_cls.__chalk_error_builder__
            builder.add_diagnostic(
                range=builder.annotation_range(t.attribute_name),
                message=(
                    f"The attribute '{t.features_cls.__name__}.{t.attribute_name}' "
                    f"has a join filter ({t.join}) but its type annotation is not a feature class or "
                    f"DataFrame ({t.typ})."
                ),
                label="Incorrect join type annotation",
                raise_error=TypeError,
                code="34",
                code_href="https://docs.chalk.ai/docs/has-many",
            )

        elif t.is_feature_time:
            feature_time_kind_gql = UpsertFeatureTimeKindGQL()

        else:
            assert t.typ is not None
            try:
                converted_type = _convert_type(t.converter.rich_type, is_windowed=t.is_windowed)
            except Exception as e:
                builder = t.features_cls.__chalk_error_builder__
                if builder:
                    builder.add_diagnostic(
                        message=e.args[0] if len(e.args) > 0 else str(e),
                        range=builder.annotation_range(t.attribute_name) or builder.property_range(t.attribute_name),
                        code="39",
                        label="invalid type",
                        code_href="https://docs.chalk.ai/docs/feature-types",
                    )
                raise e
            pretty_name = repr(t.converter.rich_type)
            if isinstance(t.converter.rich_type, type):  # pyright: ignore[reportUnnecessaryIsInstance]
                pretty_name = _parse_generic_params_recursively(t.converter.rich_type)
            pretty_name = pretty_name.replace("typing.", "")
            if t.is_windowed:
                pretty_name = f"Windowed[{pretty_name}]"

            feature_version: VersionInfo | None = t.version

            default_json = None
            if t.default is not Ellipsis:
                try:
                    default_json = json.dumps(t.converter.from_rich_to_json(t.default))
                except:
                    pass
            try:
                serialized_dtype = serialize_dtype(t.converter.pyarrow_dtype)
            except Exception as e:
                builder = t.features_cls.__chalk_error_builder__
                if builder:
                    builder.add_diagnostic(
                        message=f"Feature '{t.root_fqn}' has an invalid type: {e.args[0] if len(e.args) > 0 else str(e)}",
                        range=builder.annotation_range(t.attribute_name) or builder.property_range(t.attribute_name),
                        code="39",
                        label="invalid type",
                        code_href="https://docs.chalk.ai/docs/feature-types",
                    )
                raise e

            scalar_kind_gql = UpsertScalarKindGQL(
                scalarKind=pretty_name,
                primary=t.primary,
                baseClasses=converted_type.bases,
                version=None if feature_version is None else feature_version.version,
                versionInfo=None
                if feature_version is None
                else VersionInfoGQL(
                    version=feature_version.version,
                    maximum=feature_version.maximum,
                    default=feature_version.default,
                    versions=[f.fqn for f in feature_version.reference.values()],
                ),
                dtype=serialized_dtype,
                defaultValueJson=default_json,
                hasExplicitDtype=t._pyarrow_dtype is not None,  # pyright: ignore[reportPrivateUsage]
            )

        window_buckets = None
        if t.window_buckets is not None:
            window_buckets = sorted(list(float(x) for x in t.window_buckets))
        elif t.group_by_windowed is not None:
            window_buckets = sorted(list(float(x) for x in t.group_by_windowed.buckets_seconds))

        mat = t.window_materialization_parsed
        return UpsertFeatureGQL(
            id=UpsertFeatureIdGQL(
                fqn=t.fqn,
                name=t.name,
                namespace=t.namespace,
                isPrimary=t.primary,
                className=t.features_cls.__name__,
                attributeName=t.attribute_name,
                explicitNamespace=t.namespace != to_snake_case(t.features_cls.__name__),
            ),
            windowMaterialization=None
            if mat is None
            else UpsertWindowMaterializationGQL(
                namespace=mat.namespace,
                groupBy=[
                    UpsertFeatureReferenceGQL(
                        underlying=_get_feature_id(f, root_fqn=False),
                        path=[_get_path_component(p) for p in f.path or []],
                    )
                    for f in mat.group_by
                ],
                bucketDuration=float(mat.bucket_duration_seconds),
                aggregation=mat.aggregation,
                aggregateOn=UpsertFeatureReferenceGQL(
                    underlying=_get_feature_id(mat.aggregate_on, root_fqn=False),
                    path=[_get_path_component(p) for p in t.path or []],
                )
                if mat.aggregate_on is not None
                else None,
                dtype=serialize_dtype(mat.pyarrow_dtype),
                backfillResolver=mat.backfill_resolver,
                backfillLookbackDuration=float(mat.backfill_lookback_duration_seconds)
                if mat.backfill_lookback_duration_seconds is not None
                else None,
                backfillStartTime=mat.backfill_start_time,
                backfillSchedule=mat.backfill_schedule,
                continuousResolver=mat.continuous_resolver,
                continuousBufferDuration=float(mat.continuous_buffer_duration_seconds)
                if mat.continuous_buffer_duration_seconds is not None
                else None,
            ),
            maxStaleness=timedelta_to_duration(t.max_staleness),
            description=t.description,
            owner=t.owner,
            windowBuckets=window_buckets,
            etlOfflineToOnline=t.etl_offline_to_online,
            tags=None if t.tags is None else list(t.tags),
            hasOneKind=has_one_kind_gql,
            hasManyKind=has_many_kind_gql,
            scalarKind=scalar_kind_gql,
            featureTimeKind=feature_time_kind_gql,
            namespacePath=paths.get_classpath_or_name(t.features_cls),
            isSingleton=t.is_singleton,
        )

    if is_feature_set_class(t):
        return FeatureClassGQL(
            isSingleton=t.__chalk_is_singleton__,
            doc=t.__doc__,
            name=t.__chalk_namespace__,
            owner=t.__chalk_owner__,
            tags=t.__chalk_tags__,
        )

    if isinstance(t, Filter):
        while isinstance(t.lhs, Filter):
            t = t.lhs

        if isinstance(t.lhs, Feature):
            assert (
                t.lhs.is_scalar or t.lhs.is_feature_time
            ), f"Filters must be on scalar features, but {t.lhs.fqn} is not a scalar."

        if isinstance(t.rhs, Feature):
            assert (
                t.rhs.is_scalar or t.rhs.is_feature_time
            ), f"Filters must be on scalar features, but {t.rhs.fqn} is not a scalar."

        return UpsertFilterGQL(
            lhs=_get_feature_id(t.lhs, root_fqn=True),
            op=t.operation,
            rhs=_get_feature_id(t.rhs, root_fqn=True),
        )

    raise ValueError(f"Unable to convert {t} to GQL")


def convert_type_crontab(cron: Union[CronTab, Duration]) -> str:
    return f"{cron.total_seconds()}s" if isinstance(cron, timedelta) else cron


def gather_cdc_sources() -> List[UpsertCDCSourceGQL]:
    integration_name_to_tables = collections.defaultdict(list)
    for source in BaseSQLSource.registry:
        if isinstance(source, TableIngestMixIn):
            for schema_dot_table, preferences in source.ingested_tables.items():
                if preferences.cdc is True:
                    assert isinstance(source, BaseSQLSource)
                    integration_name_to_tables[source.name].append(schema_dot_table)
    return [
        UpsertCDCSourceGQL(
            integrationName=integration_name,
            schemaDotTableList=tables,
        )
        for integration_name, tables in integration_name_to_tables.items()
    ]



from __future__ import annotations

import collections
import sys
import time
import uuid
from contextlib import nullcontext
from datetime import datetime, timedelta
from functools import cached_property
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple

from rich import box
from rich.console import Group
from rich.padding import Padding
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    Task,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
)
from rich.style import Style
from rich.table import Table
from rich.text import Text
from tenacity import Retrying, retry_if_exception_message, stop_after_attempt
from tenacity.wait import wait_exponential_jitter

import chalk
from chalk import EnvironmentId
from chalk._reporting.models import BatchOpKind, BatchOpStatus, BatchReport
from chalk._reporting.rich.color import PASTELY_CYAN, SHADOWY_LAVENDER, UNDERLYING_CYAN
from chalk._reporting.rich.live import ChalkLive
from chalk.client import ChalkBaseException, ChalkError
from chalk.client.exc import CHALK_TRACE_ID_KEY
from chalk.utils.log_with_context import get_logger

if TYPE_CHECKING:
    from chalk.client.client_impl import ChalkAPIClientImpl

_logger = get_logger(__name__)


_MAX_ERRORS_TO_DISPLAY = 10
_TTableRow = Tuple[str, str]

try:
    from IPython import get_ipython  # noqa  # pyright: ignore

    # n.b. sometimes this is the `ipython` shell, not a notebook.
    _running_in_jupyter_notebook = get_ipython() is not None
except ImportError:
    _running_in_jupyter_notebook = False


class MofOptionalNCompleteColumn(MofNCompleteColumn):
    """
    Just like MofNCompleteColumn but renders just the completed count,
    instead of `completed / ?` if there is no total
    """

    def render(self, task: Task) -> Text:
        """Show completed/total."""
        completed = int(task.completed)
        if task.total:
            total = int(task.total)
            total_width = len(str(total))
            fmt = f"{completed:{total_width}d}{self.separator}{total}"
        else:
            fmt = f"{completed}"
        return Text(fmt, style="progress.download")


OFFLINE_QUERY = "Offline Query"


class ProgressService:
    def __init__(
        self,
        operation_id: uuid.UUID,
        client: "ChalkAPIClientImpl",
        environment_id: EnvironmentId,
        num_computers: int,
        show_progress: bool,
        caller_method: Optional[str],
    ):
        """
        :param operation_id: The ID of the operation that the ProgressService
            is polling for. This is commonly a `DatasetRevision` ID.
        :param client: A ChalkClient instance
        :param caller_method: Passed only if the caller is a Dataset method. When we call Dataset methods,
            the ProgressService is implicitly called, so we want to explain to the user that their method
            will be executed once the ProgressService finishes polling for the operation to complete.
        """
        super().__init__()
        self.operation_id = operation_id
        self.client = client
        self.caller_method = caller_method
        self._environment_id = environment_id
        self._operation_kind: Optional[BatchOpKind] = None
        self._num_computers = num_computers
        self._shard_id = 0
        if show_progress and not (_running_in_jupyter_notebook or sys.stdout.isatty()):
            _logger.warning("Progress display is not supported in dumb terminals. Progress will not be shown.")
            show_progress = False
        if show_progress:
            self.resolver_progress = Progress(
                TextColumn("  "),
                TimeElapsedColumn(),
                BarColumn(
                    finished_style=Style(color=PASTELY_CYAN),
                    style=Style(color=SHADOWY_LAVENDER),
                    complete_style=Style(color=UNDERLYING_CYAN),
                ),
                MofOptionalNCompleteColumn(),
                TextColumn("runs", style=Style(color=PASTELY_CYAN)),
                TextColumn("{task.description}"),
            )
            self.main_progress = Progress(
                TextColumn("{task.description}"),
                SpinnerColumn(style=Style(color=SHADOWY_LAVENDER)),
                TimeElapsedColumn(),
            )
            self.explainer_text = (
                Text(
                    f"The `DatasetRevision` is still being computed. {caller_method}() will execute once computation is complete.",
                    style=Style(color=PASTELY_CYAN),
                )
                if caller_method
                else None
            )
            if self._num_computers > 1:
                self.shard_progress = Progress(
                    TextColumn("{task.description}"),
                    MofNCompleteColumn(),
                    BarColumn(
                        finished_style=Style(color=PASTELY_CYAN),
                        style=Style(color=SHADOWY_LAVENDER),
                        complete_style=Style(color=UNDERLYING_CYAN),
                    ),
                    TaskProgressColumn(),
                )
            else:
                self.shard_progress = None
        else:
            self.resolver_progress = None
            self.main_progress = None
            self.explainer_text = None
            self.shard_progress = None
        self.show_progress = show_progress

    @cached_property
    def enclosure_panel(self) -> Panel | None:
        if self.main_progress is None or self.resolver_progress is None:
            return None
        if self.shard_progress is None:
            progress_bar_group = Group(self.main_progress, self.resolver_progress)
        else:
            blank_line = Padding(Text(""), pad=(0, 0, 0, 0))
            progress_bar_group = Group(self.shard_progress, blank_line, self.main_progress, self.resolver_progress)
        panel_content_table = Table.grid()
        if self.explainer_text is not None:
            panel_content_table.add_row(Padding(self.explainer_text, (0, 0, 1, 0)))
        panel_content_table.add_row(progress_bar_group)

        enclosure_panel = Panel.fit(
            panel_content_table, title="chalk", border_style=Style(color=PASTELY_CYAN), box=box.ROUNDED, padding=(1, 2)
        )

        return enclosure_panel

    def handle_resolver_update(self, batch_report: BatchReport, fqn_to_task_id: dict[str, TaskID]) -> None:
        if self.resolver_progress is None:
            return None
        for resolver_report in batch_report.resolvers:
            fqn = resolver_report.resolver_fqn
            if fqn not in fqn_to_task_id:
                short_name = fqn.split(".")[-1]
                fqn_to_task_id[fqn] = self.resolver_progress.add_task(
                    description=f"[ {short_name} ]",
                    total=resolver_report.progress.total or None,
                )

        for resolver_report in batch_report.resolvers:
            fqn = resolver_report.resolver_fqn
            resolver_task_id = fqn_to_task_id[fqn]

            rows_done_new = resolver_report.progress.computed + resolver_report.progress.failed
            self.resolver_progress.update(
                resolver_task_id,
                completed=rows_done_new,
                # total might change as the same resolver gets invoked in a separate node
                total=resolver_report.progress.total or None,
            )

    def handle_success(
        self,
        main_task_id: TaskID,
        operation_display_type: str,
        fqn_to_task_id: dict[str, TaskID],
        shard_task_id: TaskID | None,
    ):
        if shard_task_id is not None and self.shard_progress is not None:
            self.shard_progress.update(shard_task_id, advance=1)
        if self.enclosure_panel is not None:
            if self._shard_id == self._num_computers:
                self.enclosure_panel.title = "chalk ■"
        if self.main_progress is not None:
            if self._shard_id == self._num_computers:
                self.main_progress.update(main_task_id, description=f"{operation_display_type} completed", completed=1)
        if self.resolver_progress is not None:
            for task_id in fqn_to_task_id.values():
                self.resolver_progress.reset(task_id)

    def _get_failing_resolver(self, batch_report: BatchReport) -> Optional[str]:
        """
        :return: The name of the resolver that failed, or None if no resolver failed.
        """
        for resolver in batch_report.resolvers:
            if resolver.status == BatchOpStatus.FAILED:
                return resolver.resolver_fqn

    @staticmethod
    def _get_pkey_display_value(error: ChalkError) -> Optional[str]:
        if error.display_primary_key is None:
            return None

        value = ""
        if error.display_primary_key_fqn:
            value += error.display_primary_key_fqn + ": "

        value += error.display_primary_key

        return value

    def _get_error_details(self, errors: List[ChalkError]) -> List[_TTableRow]:
        if not errors:
            return [
                (
                    "Error Message",
                    "Unfortunately the cause of this error is unknown. Please contact Chalk for support and provide the Revision ID or the Trace ID above.",
                )
            ]

        truncated_from = None
        if len(errors) > _MAX_ERRORS_TO_DISPLAY:
            truncated_from = len(errors)
            errors = errors[:_MAX_ERRORS_TO_DISPLAY]

        grouped_details: Dict[int, List[_TTableRow]] = collections.defaultdict(list)
        for i, e in enumerate(errors):
            grouped_details[i].append((f"Error {i + 1}", e.message) if len(errors) > 1 else ("Error", e.message))
            if e.exception and e.exception.message:
                grouped_details[i].append((f"Exception", e.exception.message))
            if e.exception and e.exception.stacktrace:
                grouped_details[i].append((f"Stacktrace", e.exception.stacktrace))
            if val := self._get_pkey_display_value(e):
                grouped_details[i].append((f"Pkey", val))

        flattened_details = sum(list(grouped_details.values()), [])
        if truncated_from:
            flattened_details.append(
                ("Errors Truncated", f"Displaying only {_MAX_ERRORS_TO_DISPLAY} out of {truncated_from}")
            )

        return flattened_details

    def handle_and_raise_error(
        self,
        batch_report: BatchReport,
        main_task_id: TaskID,
        operation_display_type: str,
        is_fatal: bool = True,
    ) -> None:
        panel = self.enclosure_panel
        main_progress = self.main_progress
        if panel is None or main_progress is None:
            return

        all_errors = batch_report.all_errors or []
        if not all_errors and batch_report.error:
            # `all_errors` is added long after `error`.
            # We might be hitting old server code here,
            # make sure we can still talk to old server.
            all_errors.append(batch_report.error)

        progress_table = panel.renderable

        if is_fatal:
            self.handle_explainer_text_if_erring()
        else:
            self.handle_explainer_text_if_resolver_runtime_error()
        panel.border_style = Style(color="red" if is_fatal else "yellow")
        main_progress.update(
            main_task_id,
            description=f"Error occurred while executing {operation_display_type}",
        )

        key_to_display_key = {
            CHALK_TRACE_ID_KEY: "Trace ID",
        }

        metadata = batch_report.operation_metadata or {}
        metadata_tuples = [(key_to_display_key.get(k, k), v) for k, v in metadata.items()]
        failing_resolver = self._get_failing_resolver(batch_report)

        table_contents: List[_TTableRow] = [
            ("Revision ID", str(self.operation_id)),
            *metadata_tuples,
            *([("Failing Resolver", failing_resolver)] if failing_resolver else []),
            *self._get_error_details(all_errors),
        ]
        error_box = Table(
            box=box.SQUARE,
            show_lines=True,
            show_header=False,
            style=Style(color="red" if is_fatal else "yellow"),
            row_styles=[Style(color="red" if is_fatal else "yellow")] * len(table_contents),
        )

        for row in table_contents:
            error_box.add_row(*row)

        assert isinstance(progress_table, Table)
        progress_table.add_row(Padding(Text(""), (0, 0, 1, 0)))
        progress_table.add_row(error_box)

    def handle_explainer_text_if_erring(self):
        if self.explainer_text:
            assert self.caller_method is not None
            self.explainer_text.plain = (
                f"The computation for this `DatasetRevision` has failed. {self.caller_method}() cannot be executed."
            )
            self.explainer_text.style = Style(color="red")

    def handle_explainer_text_if_resolver_runtime_error(self):
        if self.explainer_text:
            assert self.caller_method is not None
            self.explainer_text.plain = f"The computation for this `DatasetRevision` has failed due to a resolver runtime error. {self.caller_method}() can still be executed."
            self.explainer_text.style = Style(color="yellow")

    def poll_report(
        self,
        timeout: float | timedelta | None,
        completion_timeout: float | timedelta | None,
    ) -> Iterable[BatchReport]:
        if completion_timeout is None:
            completion_deadline = datetime.max
        else:
            if not isinstance(completion_timeout, timedelta):
                completion_timeout = timedelta(seconds=completion_timeout)
            completion_deadline = datetime.now() + completion_timeout

        if timeout is None:
            must_receive_next_report_by = datetime.max
        else:
            if not isinstance(timeout, timedelta):
                timeout = timedelta(seconds=timeout)
            must_receive_next_report_by = datetime.now() + timeout

        while datetime.now() < must_receive_next_report_by and datetime.now() < completion_deadline:
            time.sleep(0.5)

            for attempt in Retrying(
                stop=stop_after_attempt(5),
                wait=wait_exponential_jitter(),
                reraise=True,
                retry=retry_if_exception_message(match="504|dialing"),
            ):
                with attempt:
                    batch_report = self.client.get_batch_report(self.operation_id, self._environment_id, self._shard_id)
                    if batch_report is not None:
                        if timeout is not None:
                            must_receive_next_report_by = datetime.now() + timeout
                        self._operation_kind = batch_report.operation_kind
                        yield batch_report
        else:
            if datetime.now() >= completion_deadline:
                raise TimeoutError(
                    f"Timed out waiting ({completion_timeout}) for completion of operation with id '{self.operation_id}' (chalkpy=={chalk.__version__})"
                )
            else:
                raise TimeoutError(
                    f"Timed out waiting ({timeout}) for next status report for operation with id '{self.operation_id}' (chalkpy=={chalk.__version__})"
                )

    def await_operation(self, must_fail_on_resolver_error: bool, timeout: float | timedelta | None) -> None:
        fqn_to_task_id: dict[str, TaskID] = {}
        initial_description = "Starting query execution"
        if self.main_progress is not None:
            main_task_id = self.main_progress.add_task(description=initial_description, total=1)
            if self.shard_progress is not None:
                shard_task_id = self.shard_progress.add_task(description="Processing Shards", total=self._num_computers)
            else:
                shard_task_id = None
        else:
            main_task_id = None
            shard_task_id = None
        context = ChalkLive(self.enclosure_panel, auto_refresh=True) if self.show_progress else nullcontext()
        with context:
            for batch_report in self.poll_report(timeout, completion_timeout=self.client.default_job_timeout):

                # Update main progress text
                if main_task_id is not None:
                    assert self.main_progress is not None
                    assert main_task_id is not None
                    if self.main_progress.tasks[main_task_id].description == initial_description:
                        description = f"Executing {OFFLINE_QUERY}: "
                        self.main_progress.update(main_task_id, description=description)
                self.handle_resolver_update(batch_report=batch_report, fqn_to_task_id=fqn_to_task_id)
                if batch_report.status == BatchOpStatus.COMPLETED:
                    self._shard_id += 1
                    if main_task_id is not None:
                        self.handle_success(
                            main_task_id=main_task_id,
                            operation_display_type=OFFLINE_QUERY,
                            fqn_to_task_id=fqn_to_task_id,
                            shard_task_id=shard_task_id,
                        )
                    if self._shard_id == self._num_computers:
                        break

                if batch_report.status == BatchOpStatus.FAILED:
                    if main_task_id is not None:
                        if all(e.is_resolver_runtime_error() for e in (batch_report.all_errors or ())):
                            # There is no chalk error, we should NOT raise an exception, but definitely warn the user
                            self.handle_and_raise_error(
                                batch_report=batch_report,
                                main_task_id=main_task_id,
                                operation_display_type=OFFLINE_QUERY,
                                is_fatal=must_fail_on_resolver_error,
                            )
                        else:
                            self.handle_and_raise_error(
                                batch_report=batch_report,
                                main_task_id=main_task_id,
                                operation_display_type=OFFLINE_QUERY,
                            )
                    if must_fail_on_resolver_error:
                        raise ChalkBaseException(
                            trace_id=(batch_report.operation_metadata or {}).get(CHALK_TRACE_ID_KEY),
                            errors=batch_report.all_errors,
                        )
                    break  # should have raised but just in case



import inspect
from typing import Any, Callable, Dict, Literal, Optional, TypeVar, Union

from typing_extensions import TYPE_CHECKING, ParamSpec

from chalk._lsp.error_builder import get_resolver_error_builder
from chalk.features.tag import Environments
from chalk.state import KeyedState
from chalk.streams._kafka_source import KafkaSource
from chalk.streams._kinesis_source import KinesisSource
from chalk.streams._pubsub_source import PubSubSource
from chalk.streams._windows import (
    MaterializationWindowConfig,
    Windowed,
    get_name_with_duration,
    group_by_windowed,
    windowed,
)
from chalk.streams.base import StreamSource
from chalk.utils import MachineType

if TYPE_CHECKING:
    from chalk.features.resolver import ResolverProtocol, StreamResolver

__all__ = (
    "KafkaSource",
    "KeyedState",
    "KinesisSource",
    "MaterializationWindowConfig",
    "PubSubSource",
    "StreamSource",
    "Windowed",
    "get_name_with_duration",
    "group_by_windowed",
    "stream",
    "windowed",
)

P = ParamSpec("P")
T = TypeVar("T")
V = TypeVar("V")


def stream(
    *,
    source: StreamSource,
    mode: Optional[Literal["continuous", "tumbling"]] = None,
    environment: Optional[Environments] = None,
    machine_type: Optional[MachineType] = None,
    owner: Optional[str] = None,
    parse: Optional[Callable[[T], V]] = None,
    keys: Optional[Dict[str, Any]] = None,
    timestamp: Optional[str] = None,
    updates_materialized_aggregations: bool = True,
) -> Union[Callable[[Callable[P, T]], "ResolverProtocol[P, T]"], "ResolverProtocol[P, T]"]:
    """Decorator to create a stream resolver.

    Parameters
    ----------
    source
        The streaming source, e.g. `KafkaSource(...)` or `KinesisSource(...)` or `PubSubSource(...)`.
    mode
        This parameter is defined when the streaming resolver returns a windowed feature.
        Tumbling windows are fixed-size, contiguous and non-overlapping time intervals. You can think of
        tumbling windows as adjacently arranged bins of equal width.
        Tumbling windows are most often used alongside `max_staleness` to allow the features
        to be sent to the online store and offline store after each window period.

        Continuous windows, unlike tumbling window, are overlapping and exact.
        When you request the value of a continuous window feature, Chalk looks
        at all the messages received in the window and computes the value on-demand.

        See more at https://docs.chalk.ai/docs/windowed-streaming#window-modes
    parse
        A callable that will interpret an input prior to the invocation of the resolver.
        Parse functions can serve many functions, including pre-parsing bytes,
        skipping unrelated messages, or supporting rekeying.

        See more at https://docs.chalk.ai/docs/streams#parsing
    keys
        A mapping from input `BaseModel` attribute to Chalk feature attribute to support continuous streaming re-keying.
        This parameter is required for continuous resolvers.
        Features that are included here do not have to be explicitly returned in the stream resolver:
        the feature will automatically be set to the key value used for aggregation.
    timestamp
        An optional string specifying an input attribute as the timestamp used for windowed aggregations.
    updates_materialized_aggregations
        If set to `False`, the stream resolver will not update materialized aggregations, but is still eligible for ETL.

    Other Parameters
    ----------------
    environment
        Environments are used to trigger behavior in different deployments
        such as staging, production, and local development.

        Environment can take one of three types:
            - `None` (default) - candidate to run in every environment
            - `str` - run only in this environment
            - `list[str]` - run in any of the specified environment and no others

        Read more at https://docs.chalk.ai/docs/resolver-environments
    machine_type
        You can optionally specify that resolvers need to run
        on a machine other than the default. Must be configured
        in your deployment.
    owner
        Individual or team responsible for this resolver.
        The Chalk Dashboard will display this field, and alerts
        can be routed to owners.

    Returns
    -------
    Callable[[Any, ...], Any]
        A callable function! You can unit-test stream resolvers as you would
        unit-test any other code.
    """
    frame = inspect.currentframe()
    assert frame is not None
    caller_frame = frame.f_back
    assert caller_frame is not None
    caller_globals = caller_frame.f_globals
    caller_locals = caller_frame.f_locals
    del frame
    from chalk.features.resolver import parse_and_register_stream_resolver

    def decorator(fn: Callable[P, T]) -> "StreamResolver[P,T]":
        caller_filename = inspect.getsourcefile(fn) or "unknown_file"
        error_builder = get_resolver_error_builder(fn)
        return parse_and_register_stream_resolver(
            caller_globals=caller_globals,
            caller_locals=caller_locals,
            fn=fn,
            source=source,
            mode=mode,
            caller_filename=caller_filename,
            caller_line=caller_frame.f_lineno,
            error_builder=error_builder,
            environment=environment,
            machine_type=machine_type,
            message=None,
            owner=owner,
            parse=parse,
            keys=keys,
            timestamp=timestamp,
            updates_materialized_aggregations=updates_materialized_aggregations,
        )

    return decorator



from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING, Any, AsyncIterable, Dict, Iterable, Mapping, Optional

import pyarrow as pa

from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sql._internal.sql_source import BaseSQLSource, SQLSourceKind
from chalk.sql.protocols import ChalkQueryProtocol, StringChalkQueryProtocol
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine import Engine
    from sqlalchemy.engine.url import URL


_logger = get_logger(__name__)


_DYNAMODB_AWS_CLIENT_ID_OVERRIDE_NAME = "DYNAMODB_AWS_CLIENT_ID_OVERRIDE"
_DYNAMODB_AWS_CLIENT_SECRET_OVERRIDE_NAME = "DYNAMODB_AWS_CLIENT_SECRET_OVERRIDE"
_DYNAMODB_AWS_ROLE_ARN_OVERRIDE_NAME = "DYNAMODB_AWS_ROLE_ARN_OVERRIDE"
_DYNAMODB_AWS_REGION_OVERRIDE_NAME = "DYNAMODB_AWS_REGION_OVERRIDE"
_DYNAMODB_ENDPOINT_OVERRIDE_NAME = "DYNAMODB_ENDPOINT_OVERRIDE"
_DYNAMODB_ENGINE_ARGUMENTS_NAME = "DYNAMODB_ENGINE_ARGUMENTS"


class DynamoDBSourceImpl(BaseSQLSource):
    kind = SQLSourceKind.dynamodb

    def __init__(
        self,
        name: str | None = None,
        aws_client_id_override: str | None = None,
        aws_client_secret_override: str | None = None,
        aws_role_arn_override: str | None = None,
        aws_region_override: str | None = None,
        endpoint_override: str | None = None,
        engine_args: Dict[str, Any] | None = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        try:
            import boto3
            import pydynamodb
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[dynamodb]")
        del boto3, pydynamodb  # unused

        self.ingested_tables: Dict[str, Any] = {}
        self.aws_client_id_override = aws_client_id_override or load_integration_variable(
            integration_name=name, name=_DYNAMODB_AWS_CLIENT_ID_OVERRIDE_NAME, override=integration_variable_override
        )
        self.aws_client_secret_override = aws_client_secret_override or load_integration_variable(
            integration_name=name,
            name=_DYNAMODB_AWS_CLIENT_SECRET_OVERRIDE_NAME,
            override=integration_variable_override,
        )
        self.aws_role_arn_override = aws_role_arn_override or load_integration_variable(
            integration_name=name, name=_DYNAMODB_AWS_ROLE_ARN_OVERRIDE_NAME, override=integration_variable_override
        )
        self.aws_region_override = aws_region_override or load_integration_variable(
            integration_name=name, name=_DYNAMODB_AWS_REGION_OVERRIDE_NAME, override=integration_variable_override
        )
        self.endpoint_override = endpoint_override or load_integration_variable(
            integration_name=name, name=_DYNAMODB_ENDPOINT_OVERRIDE_NAME, override=integration_variable_override
        )
        if self.aws_role_arn_override is not None and (
            self.aws_client_id_override is not None or self.aws_client_secret_override is not None
        ):
            raise ValueError(
                "If you provide an AWS role ARN override, you cannot provide an AWS client ID or secret override."
            )
        if self.aws_client_id_override is not None and self.aws_client_secret_override is None:
            raise ValueError(
                "If you provide an AWS client ID override, you must also provide an AWS client secret override."
            )
        if self.aws_client_secret_override is not None and self.aws_client_id_override is None:
            raise ValueError(
                "If you provide an AWS client secret override, you must also provide an AWS client ID override."
            )
        eargs = {}
        if engine_args is not None:
            eargs = engine_args
        else:
            found = load_integration_variable(
                integration_name=name,
                name=_DYNAMODB_ENGINE_ARGUMENTS_NAME,
                override=integration_variable_override,
                parser=json.loads,
            )
            if found is not None:
                eargs = found

        # The SQL sources are read-only, so transactions are not needed.
        # Setting the isolation level on the engine, instead of the connection, avoids
        # a DBAPI statement to reset the transactional level back to the default before returning the
        # connection to the pool.
        eargs.setdefault(
            "isolation_level",
            os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"),
        )

        # # Dynamo defaults to ReadWrite, but Chalk only reads data from dynamo.
        # eargs.setdefault("read_only", True)

        BaseSQLSource.__init__(self, name=name, engine_args=eargs, async_engine_args={})

    def get_engine(self) -> "Engine":
        try:
            from pydynamodb.sqlalchemy_dynamodb.pydynamodb import DynamoDBDialect
            from sqlalchemy.engine import create_engine
        except ImportError:
            raise missing_dependency_exception("chalkpy[dynamodb]")

        if self._engine is None:  # pyright: ignore[reportUnnecessaryComparison]
            self._engine = create_engine(
                self.local_engine_url().render_as_string(), executor=DynamoDBDialect()
            ).execution_options(**self.engine_args)

        return self._engine

    def local_engine_url(self) -> "URL":
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="dynamodb",
            host=self.endpoint_override
            or (f"dynamodb.{self.aws_region_override}.amazonaws.com" if self.aws_region_override else None),
            username=self.aws_client_id_override,
            password=self.aws_client_secret_override,
        )

    def execute_query(self, *args: Any, **kwargs: Any) -> Iterable[pa.RecordBatch]:
        raise NotImplementedError("DynamoDB sources cannot be queried without native sql enabled")

    async def async_execute_query(self, *args: Any, **kwargs: Any) -> AsyncIterable[pa.RecordBatch]:
        raise NotImplementedError("DynamoDB sources cannot be queried without native sql enabled")
        yield  # noqa

    def query(self, *args: Any, **kwargs: Any) -> ChalkQueryProtocol:
        raise NotImplementedError("DynamoDB sources cannot be queried without native sql enabled")

    def query_sql_file(self, *args: Any, **kwargs: Any) -> StringChalkQueryProtocol:
        raise NotImplementedError("DynamoDB sources cannot be queried without native sql enabled")

    def query_string(self, *args: Any, **kwargs: Any) -> StringChalkQueryProtocol:
        raise NotImplementedError(
            "DynamoDB sources cannot be queried in python resolvers directly. Create a sql file resolver and enable native sql to query this source."
        )

    def _execute_query_inefficient(self, *args: Any, **kwargs: Any) -> pa.RecordBatch:
        raise NotImplementedError("DynamoDB sources cannot be queried without native sql enabled")

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(
                    _DYNAMODB_AWS_CLIENT_ID_OVERRIDE_NAME, self.name, self.aws_client_id_override
                ),
                create_integration_variable(
                    _DYNAMODB_AWS_CLIENT_SECRET_OVERRIDE_NAME, self.name, self.aws_client_secret_override
                ),
                create_integration_variable(
                    _DYNAMODB_AWS_ROLE_ARN_OVERRIDE_NAME, self.name, self.aws_role_arn_override
                ),
                create_integration_variable(_DYNAMODB_AWS_REGION_OVERRIDE_NAME, self.name, self.aws_region_override),
                create_integration_variable(_DYNAMODB_ENDPOINT_OVERRIDE_NAME, self.name, self.endpoint_override),
            ]
            if v is not None
        }

    # def get_sqlglot_dialect(self) -> str | None:
    #     return "partiql"



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.common.v1 import chalk_error_pb2 as _chalk_error_pb2
from chalk._gen.chalk.common.v1 import dataset_response_pb2 as _dataset_response_pb2
from chalk._gen.chalk.common.v1 import offline_query_pb2 as _offline_query_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class CreateOfflineQueryRequest(_message.Message):
    __slots__ = ("request",)
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    request: _offline_query_pb2.OfflineQueryRequest
    def __init__(self, request: _Optional[_Union[_offline_query_pb2.OfflineQueryRequest, _Mapping]] = ...) -> None: ...

class CreateOfflineQueryResponse(_message.Message):
    __slots__ = ("is_finished", "version", "urls", "errors", "columns", "dataset_response")
    IS_FINISHED_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    URLS_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    DATASET_RESPONSE_FIELD_NUMBER: _ClassVar[int]
    is_finished: bool
    version: int
    urls: _containers.RepeatedScalarFieldContainer[str]
    errors: _containers.RepeatedCompositeFieldContainer[_chalk_error_pb2.ChalkError]
    columns: _offline_query_pb2.ColumnMetadataList
    dataset_response: _dataset_response_pb2.DatasetResponse
    def __init__(
        self,
        is_finished: bool = ...,
        version: _Optional[int] = ...,
        urls: _Optional[_Iterable[str]] = ...,
        errors: _Optional[_Iterable[_Union[_chalk_error_pb2.ChalkError, _Mapping]]] = ...,
        columns: _Optional[_Union[_offline_query_pb2.ColumnMetadataList, _Mapping]] = ...,
        dataset_response: _Optional[_Union[_dataset_response_pb2.DatasetResponse, _Mapping]] = ...,
    ) -> None: ...



import ast
import inspect
import textwrap
from typing import Any, Callable, Dict, Optional

from chalk._lsp.error_builder import ResolverErrorBuilder
from chalk.df.ast_parser import convert_slice, eval_converted_expr
from chalk.utils.collections import get_unique_item


class ResolverAnnotationParser:
    def __init__(
        self,
        resolver: Callable,
        glbs: Optional[Dict[str, Any]],
        lcls: Optional[Dict[str, Any]],
        error_builder: ResolverErrorBuilder,
    ):
        super().__init__()
        self.resolver = resolver
        self.glbs = glbs
        self.lcls = lcls

        self._args = {arg.arg: arg for arg in self._get_resolver_args()}
        self.builder = error_builder

    def _get_resolver_args(self):
        source = inspect.getsource(self.resolver)
        parsed_source = ast.parse(textwrap.dedent(source))
        function_def = get_unique_item(parsed_source.body)
        if not isinstance(function_def, (ast.FunctionDef, ast.AsyncFunctionDef)):
            raise TypeError(f"The resolver must be a function. Received:\n\n{source}")
        args = function_def.args
        return [*args.posonlyargs, *args.args, *args.kwonlyargs]

    def parse_annotation(self, name: str):
        arg = self._args[name]
        annotation = arg.annotation
        if annotation is None:
            self.builder.add_diagnostic(
                message=(
                    f"Argument '{name}' for resolver '{self.resolver.__name__}' was not defined with a type annotation."
                ),
                code="84",
                label="missing annotation",
                range=self.builder.function_arg_value_by_name(name),
                raise_error=TypeError,
            )
        if isinstance(annotation, ast.Constant):
            val = annotation.value
            if not isinstance(val, str):
                self.builder.add_diagnostic(
                    message=f"Argument {name} has a Literal[...] type annotation, but it is not a string.",
                    code="85",
                    label="invalid annotation",
                    range=self.builder.function_arg_annotation_by_name(name),
                    raise_error=TypeError,
                )

            # string of type annotation
            val = ast.parse(val, mode="eval")
            if not isinstance(val, ast.Expr):
                self.builder.add_diagnostic(
                    message=f"Argument {name} has a Literal type annotation, but it is not a string.",
                    code="86",
                    label="invalid argument",
                    range=self.builder.function_arg_annotation_by_name(name),
                    raise_error=TypeError,
                )
            annotation = val.body

        if isinstance(annotation, ast.Subscript):
            # All fancy ast parsing would appear within the subscript of a df __getitem__
            annotation = ast.Subscript(
                value=annotation.value,
                slice=convert_slice(annotation.slice),  # pyright: ignore[reportArgumentType]
                ctx=annotation.ctx,
            )
        return eval_converted_expr(annotation, self.glbs, self.lcls)



from __future__ import annotations

import os
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Any, Mapping, Optional

import yaml

from chalk.config._validator import Validator
from chalk.config.project_config import load_project_config
from chalk.utils.log_with_context import get_logger

_logger = get_logger(__name__)


@dataclass
class TokenConfig:
    clientId: str
    clientSecret: str
    name: Optional[str] = None
    apiServer: Optional[str] = None
    activeEnvironment: Optional[str] = None

    @staticmethod
    def from_py(raw: Any, path: str) -> "TokenConfig":
        raw = Validator.dict_with_str_keys(raw, name=path)
        prefix = f"tokens.{path}"
        return TokenConfig(
            clientId=Validator.string(raw.get("clientId"), f"{prefix}.clientId"),
            clientSecret=Validator.string(raw.get("clientSecret"), f"{prefix}.clientSecret"),
            name=Validator.optional_string(raw.get("name"), f"{prefix}.name"),
            apiServer=Validator.optional_string(raw.get("apiServer"), f"{prefix}.apiServer"),
            activeEnvironment=Validator.optional_string(raw.get("activeEnvironment"), f"{prefix}.activeEnvironment"),
        )


@dataclass
class AuthConfig:
    tokens: Optional[Mapping[str, TokenConfig]]

    @staticmethod
    def from_py(raw: Any) -> "AuthConfig":
        value = Validator.dict_with_str_keys_or_none(raw, name="auth config")
        if value is None:
            return AuthConfig(tokens=None)
        tokens = Validator.dict_with_str_keys_or_none(value.get("tokens", None), "tokens")
        if tokens is None:
            return AuthConfig(tokens=None)
        return AuthConfig(tokens={k: TokenConfig.from_py(v, path=k) for k, v in tokens.items()})


def _load_global_config_uncached() -> AuthConfig | None:
    path = None
    p = os.path.join(os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~")), ".chalk.yml")
    if os.path.exists(p):
        path = p

    if path is None:
        return None

    with open(path, "r") as f:
        parsed = yaml.safe_load(f)
    return AuthConfig.from_py(parsed)


# using lru_cache for 3.8 compat
@lru_cache(maxsize=None)
def _load_global_config() -> AuthConfig | None:
    return _load_global_config_uncached()


def _en(current_env_name: str, old_env_name: str | None = None) -> str | None:
    old_env_name = old_env_name or f"_{current_env_name}"
    # Backwards compatability for old env vars.
    deprecated = os.getenv(old_env_name)
    current = os.getenv(current_env_name)
    if deprecated and current:
        _logger.warning(
            f"Using both deprecated '{old_env_name}' and new '{current_env_name}' environment variables. Choosing '{current_env_name}'"
        )
    return current or deprecated


def load_token(
    client_id: Optional[str],
    client_secret: Optional[str],
    active_environment: Optional[str],
    api_server: Optional[str],
    skip_cache: bool = False,
) -> TokenConfig | None:
    token_name = "Explicit" if client_id and client_secret else "Environment"

    client_id = client_id or _en("CHALK_CLIENT_ID")
    client_secret = client_secret or _en("CHALK_CLIENT_SECRET")
    api_server = api_server or _en("CHALK_API_SERVER")
    active_environment = active_environment or _en("CHALK_ENVIRONMENT", "_CHALK_ACTIVE_ENVIRONMENT")

    if client_id is not None and client_secret is not None:
        return TokenConfig(
            name=token_name,
            clientId=client_id,
            clientSecret=client_secret,
            apiServer=api_server or "https://api.chalk.ai",
            activeEnvironment=active_environment,
        )

    global_cfg = _load_global_config_uncached() if skip_cache else _load_global_config()
    if global_cfg is None:
        return None

    absdir = Path(os.getcwd()).absolute().resolve()
    project_config = load_project_config()
    if project_config:
        absdir = Path(project_config.local_path).parent
    tokens: Mapping[str, TokenConfig] = global_cfg.tokens or {}
    token_config = tokens.get(str(absdir)) or tokens.get("default")
    if token_config is None:
        return None

    return TokenConfig(
        name=token_config.name,
        clientId=client_id or token_config.clientId,
        clientSecret=client_secret or token_config.clientSecret,
        apiServer=api_server or token_config.apiServer or "https://api.chalk.ai",
        activeEnvironment=active_environment or token_config.activeEnvironment,
    )



from typing import Literal

from typing_extensions import TypeGuard

Operation = Literal[
    "!=",
    "<",
    "<=",
    "==",
    ">",
    ">=",
    "and",
    "in",
    "is_near_cos",
    "is_near_ip",
    "is_near_l2",
    "not in",
    "not",
    "or",
]


def is_valid_operation(operation: str) -> TypeGuard[Operation]:
    return operation in (
        "!=",
        "<",
        "<=",
        "==",
        ">",
        ">=",
        "and",
        "in",
        "is_near_cos",
        "is_near_ip",
        "is_near_l2",
        "not in",
        "not",
        "or",
    )



# pyright: reportPrivateUsage=false
from __future__ import annotations

import importlib
import importlib.util
import linecache
import os
import sys
import time
import traceback
from contextvars import ContextVar
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple, Type, Union, cast, get_args, get_origin

import pyarrow as pa

from chalk._lsp.error_builder import DiagnosticBuilder, LSPErrorBuilder
from chalk.features import Feature, Features, FeatureSetBase, Filter, unwrap_feature
from chalk.features.feature_field import WindowConfigResolved
from chalk.features.pseudofeatures import Now

if TYPE_CHECKING:
    from chalk.features.feature_set import FeaturesProtocol

from chalk.features.resolver import RESOLVER_REGISTRY, Resolver
from chalk.features.underscore import (
    Underscore,
    UnderscoreAttr,
    UnderscoreCall,
    UnderscoreFunction,
    UnderscoreItem,
    UnderscoreRoot,
)
from chalk.gitignore.helper import IgnoreConfig, get_default_combined_ignore_config
from chalk.parsed.duplicate_input_gql import (
    DiagnosticGQL,
    DiagnosticSeverityGQL,
    FailedImport,
    PositionGQL,
    PublishDiagnosticsParams,
    RangeGQL,
)
from chalk.sql import SQLSourceGroup
from chalk.sql._internal.sql_file_resolver import get_sql_file_resolvers, get_sql_file_resolvers_from_paths
from chalk.sql._internal.sql_source import BaseSQLSource
from chalk.utils.collections import ensure_tuple
from chalk.utils.duration import parse_chalk_duration_s, timedelta_to_duration
from chalk.utils.import_utils import py_path_to_module
from chalk.utils.log_with_context import get_logger
from chalk.utils.paths import get_directory_root

try:
    from types import UnionType
except ImportError:
    UnionType = Union

_logger = get_logger(__name__)
_import_logger = get_logger("chalk.import_logger")


has_imported_all_files = False


def import_all_files_once(
    file_allowlist: Optional[List[str]] = None,
    project_root: Optional[Path] = None,
    only_sql_files: bool = False,
    override: bool = False,
) -> List[FailedImport]:
    global has_imported_all_files
    if has_imported_all_files:
        return []
    failed = import_all_files(
        file_allowlist=file_allowlist,
        project_root=project_root,
        only_sql_files=only_sql_files,
        override=override,
    )
    has_imported_all_files = True
    return failed


supported_aggs = (
    "approx_count_distinct",
    "approx_percentile",
    "array_agg",
    "count",
    "max",
    "mean",
    "min",
    "std",
    "std_sample",
    "stddev",
    "stddev_sample",
    "sum",
    "var",
    "var_sample",
)


class ChalkParseError(ValueError):
    pass


def _try_parse_datetime(
    name: str,
    dt: datetime | str | None,
) -> datetime | None:
    if dt is None:
        return None
    if isinstance(dt, datetime):
        return dt
    if isinstance(dt, str):  # pyright: ignore[reportUnnecessaryIsInstance]
        try:
            return datetime.fromisoformat(dt)
        except ValueError as e:
            raise ChalkParseError(f"parsing {name}: {str(e)}")
    raise ChalkParseError(f"{name} must be a string or a datetime object. Found: {dt}")


def _try_parse_duration(
    name: str,
    d: timedelta | str | int | None,
) -> int | None:
    if d is None:
        return None
    try:
        return parse_chalk_duration_s(d)
    except ValueError as e:
        raise ChalkParseError(f"parsing {name}: {str(e)}")


def _try_parse_resolver_fqn(name: str, resolver: Any) -> str | None:
    if resolver is None:
        return None
    if isinstance(resolver, Resolver):
        return resolver.fqn
    if isinstance(resolver, str):
        return resolver
    raise ChalkParseError(
        f"{name} must be a string or a Resolver object. Found: {resolver}",
    )


def _check_types(
    feature_name: str,
    aggregation: str,
    joined_feature: Feature | None,
    this_annotation: Any,
) -> None:
    if joined_feature is None:
        if aggregation != "count":
            raise ChalkParseError(
                f"feature '{feature_name}' does not aggregate a child feature; expected 'count' aggregation"
            )
        _validate_types(
            annotation=this_annotation,
            permitted_types=(int, float),
            aggregation_name=aggregation,
            joined=False,
            feature_name=feature_name,
        )
        return

    joined_annotation = joined_feature.typ.parsed_annotation
    if aggregation not in {"count", "approx_count_distinct", "array_agg"}:
        _validate_types(
            annotation=joined_annotation,
            permitted_types=(int, float),
            aggregation_name=aggregation,
            joined=True,
            feature_name=joined_feature.name,
        )

    if aggregation in (
        "approx_percentile",
        "mean",
        "std",
        "std_sample",
        "stddev",
        "stddev_sample",
        "var",
        "var_sample",
    ):
        _validate_types(
            annotation=this_annotation,
            permitted_types=(float,),
            aggregation_name=aggregation,
            joined=False,
            feature_name=feature_name,
        )

    elif aggregation == "min" or aggregation == "max":
        if _get_underlying_type(this_annotation, feature_name) != _get_underlying_type(
            joined_annotation, joined_feature.name
        ):
            raise ChalkParseError(
                (
                    f"feature '{feature_name}' of type '{_get_printable_name(this_annotation)}' should match "
                    f"joined feature '{joined_feature.name}' of type '{_get_printable_name(joined_annotation)}' "
                    f"for '{aggregation}' aggregation"
                )
            )

    elif aggregation == "count":
        _validate_types(
            annotation=this_annotation,
            permitted_types=(int, float),
            aggregation_name=aggregation,
            joined=False,
            feature_name=feature_name,
        )

    elif aggregation == "sum":
        _validate_types(
            annotation=this_annotation,
            permitted_types=(int, float),
            aggregation_name=aggregation,
            joined=False,
            feature_name=feature_name,
        )
        if (
            _get_underlying_type(joined_annotation, feature_name) is float
            and _get_underlying_type(this_annotation, feature_name) is int
        ):
            raise ChalkParseError(
                f"feature '{feature_name}' should be a 'float' for 'sum' aggregation with joined feature '{joined_feature.name}'; found 'int'"
            )


def _validate_types(
    annotation: type, permitted_types: tuple[type, ...], aggregation_name: str, joined: bool, feature_name: str
):
    error_message = (
        f"feature '{feature_name}' should be of type {' or '.join(p.__name__ for p in permitted_types)} "
        + f"for aggregation '{aggregation_name}'; found '{_get_printable_name(annotation)}'"
    )
    if joined:
        error_message = "joined " + error_message
    t = _get_underlying_type(annotation, feature_name)
    if t not in permitted_types:
        raise ChalkParseError(error_message)


def _get_underlying_type(t: type, feature_name: str) -> type:
    if get_origin(t) in (Union, UnionType):
        args = get_args(t)
        return_type = None
        for arg in args:
            if arg == type(None):
                continue
            if return_type is not None:
                raise ChalkParseError(
                    f"feature {feature_name} is a Union over types {(str(s) for s in args)}, which includes "
                    + f"contradictory types '{str(return_type)}' and '{str(arg)}'"
                )
            return_type = arg
        return return_type if return_type is not None else t
    return t


def _parse_agg_function_call(expr: Underscore | None) -> Tuple[str, Underscore]:
    if not isinstance(expr, UnderscoreCall):
        raise ChalkParseError(
            "missing aggregation function call for materialized aggregate feature -- if materialization is enabled, the expression must include an aggregation function (e.g. .count())"
        )

    call_expr = expr
    if len(call_expr._chalk__args) > 0 or len(call_expr._chalk__kwargs) > 0:
        raise ChalkParseError("should not have any arguments or keyword arguments")

    function_attribute = call_expr._chalk__parent
    if not isinstance(function_attribute, UnderscoreAttr):
        raise ChalkParseError("expected an aggregation, like `.sum()`")

    aggregation = function_attribute._chalk__attr
    if aggregation not in supported_aggs:
        raise ChalkParseError(f"aggregation should be one of {', '.join(supported_aggs)}")

    return aggregation, function_attribute._chalk__parent


def _parse_projection(expr: Underscore | None) -> str:
    # TODO: get in here and find the filters
    if not isinstance(expr, UnderscoreAttr):
        raise ChalkParseError("expected a feature, like `_.amount`")

    attr = expr._chalk__attr
    if not isinstance(expr._chalk__parent, UnderscoreRoot):
        raise ChalkParseError("expected a single feature, like `_.amount`")

    return attr


def _get_has_many_class(
    parent: Type[Features],
    has_many_feature_name: str,
    group_names: list[str],
    aggregated_feature_name: str | None,
) -> Tuple[Type[Features], list[Feature], Feature | None]:
    """
    If the has-many class and aggregated features are found:
        Tuple[
            Type[Features], -- joined_class
            list[Feature],  -- the group_names translated to features on joined_class. Also includes the foreign join key
            Feature | None, -- the aggregated feature on joined_class, from aggregated_feature_name
            None            -- the error is None
        ]
    """
    try:
        has_many_feature = getattr(parent, has_many_feature_name)
    except Exception:
        raise ChalkParseError(f"could not find feature '{has_many_feature_name}' in the child namespace")

    underlying = unwrap_feature(has_many_feature, raise_error=False)

    if not isinstance(underlying, Feature):
        raise ChalkParseError(f"the attribute '{has_many_feature_name}' is not a feature")

    if not underlying.is_has_many:
        raise ChalkParseError(f"the attribute '{has_many_feature_name}' is not a has-many feature")

    joined_class = underlying.joined_class
    if joined_class is None:
        raise ChalkParseError(f"has-many feature '{has_many_feature_name}' is not joined to a class")

    foreign_join_key = underlying.foreign_join_key
    if foreign_join_key is None:
        raise ChalkParseError(f"has-many feature '{has_many_feature_name}' is missing a foreign join key")

    group_by_features: list[Feature] = [foreign_join_key]
    for group_name in group_names:
        joined_feature_wrapper = getattr(joined_class, group_name, None)
        if joined_feature_wrapper is None:
            raise ChalkParseError(f"joined class '{_get_printable_name(joined_class)}' missing feature '{group_name}'")

        joined_feature = unwrap_feature(joined_feature_wrapper)
        if not joined_feature.is_scalar:
            raise ChalkParseError(
                f"group feature '{joined_feature.window_stem}' not a scalar, like a 'float' or an 'int'"
            )
        group_by_features.append(joined_feature)

    if aggregated_feature_name is None:
        return joined_class, group_by_features, None

    aggregated_feature_wrapper = getattr(joined_class, aggregated_feature_name, None)
    if aggregated_feature_wrapper is None:
        raise ChalkParseError(
            f"joined class '{_get_printable_name(joined_class)}' missing feature '{aggregated_feature_name}'"
        )

    aggregated_feature = unwrap_feature(aggregated_feature_wrapper)
    if not aggregated_feature.is_scalar:
        raise ChalkParseError(
            f"joined feature '{aggregated_feature.window_stem}' not a scalar, like a 'float' or an 'int'"
        )

    return joined_class, group_by_features, aggregated_feature


def run_post_import_fixups():
    for _, fsb in FeatureSetBase.registry.items():
        for f in fsb.__chalk_group_by_materialized_windows__:
            # Fix up group by window materializations like .group_by(...).agg(...):
            #   num_theorem_lines_by_author: DataFrame = group_by_windowed(
            #       "1m", "2m", materialization={...},
            #       expression=_.theorems.group_by(_.author).agg(_.num_lines.sum()),
            #   )
            try:
                mat = parse_grouped_window(f=f)
                gbw = f.group_by_windowed
                assert gbw is not None
                gbw._window_materialization_parsed = mat
                f.window_materialization_parsed = mat
            except ChalkParseError as e:
                error = e.args[0]
                lsp_error_builder = fsb.__chalk_error_builder__
                lsp_error_builder.add_diagnostic(
                    message=f"The materialized feature '{f.name}' is incorrectly configured.",
                    label=error,
                    range=lsp_error_builder.property_value_kwarg_range(
                        feature_name=f.attribute_name,
                        kwarg="expression",
                    )
                    or lsp_error_builder.property_range(feature_name=f.attribute_name),
                    code="42",
                )

        for f in fsb.__chalk_materialized_windows__:
            # Fix up materialized windows without group by
            #   sum_spending: DataFrame = group_by_windowed(
            #       "1m", "2m", materialization={...},
            #       expression=_.transactions[_.amount].sum(),
            #   )
            assert f.underscore_expression is not None
            assert f.window_materialization is not None

            try:
                f.window_materialization_parsed = parse_windowed_materialization(f=f)
            except ChalkParseError as e:
                error = e.args[0]
                f.lsp_error_builder.add_diagnostic(
                    message=f"The windowed materialization on feature '{f.window_stem}' is incorrectly configured.",
                    label=error,
                    range=f.lsp_error_builder.property_value_kwarg_range(
                        feature_name=f.window_stem,
                        kwarg="expression",
                    )
                    or f.lsp_error_builder.property_range(feature_name=f.window_stem),
                    code="39",
                )


def parse_grouped_window(f: Feature) -> WindowConfigResolved:
    """
    _.transactions.group_by(_.mcc).agg(_.amount.sum())
    """
    assert f.features_cls is not None
    expr = f.underscore_expression
    if not isinstance(expr, UnderscoreCall):
        raise ChalkParseError("group expression should end with `.agg(...)`")
    pyarrow_dtype = f.converter.pyarrow_dtype
    kind = f.typ.parsed_annotation

    assert f.group_by_windowed is not None

    materialization = f.group_by_windowed._materialization
    bucket_duration_str = (
        materialization.get("bucket_duration", None) if isinstance(materialization, dict) else f.window_duration
    )
    assert bucket_duration_str is not None
    bucket_duration_seconds = parse_chalk_duration_s(bucket_duration_str)

    bucket_start = datetime.fromtimestamp(0, tz=timezone.utc)
    if isinstance(materialization, dict):
        bucket_start = materialization.get("bucket_start", bucket_start)

    call_expr = expr
    call_exp_parent = call_expr._chalk__parent
    if not isinstance(call_exp_parent, UnderscoreAttr):
        raise ChalkParseError("expected a `.agg(...) call")

    agg_function_name = call_exp_parent._chalk__attr
    if agg_function_name != "agg":
        raise ChalkParseError(f"expected a `.agg(...) call, but found the function call `{agg_function_name}`")

    if len(call_expr._chalk__kwargs):
        raise ChalkParseError("aggregations should not be supplied as keyword arguments")

    if len(call_expr._chalk__args) != 1:
        raise ChalkParseError("expected a single aggregation function")

    aggregation_expr = call_expr._chalk__args[0]

    aggregation, par = _parse_agg_function_call(aggregation_expr)

    # If it's an error, we'll tolerate `.count()`, so leave it be for now.
    try:
        child_feature_name = _parse_projection(par)
    except ChalkParseError:
        child_feature_name = None
    if not isinstance(par, UnderscoreAttr) and not isinstance(par, UnderscoreRoot):
        raise ChalkParseError("expected a feature, like `_.amount`")

    group_by_call = call_exp_parent._chalk__parent
    if not isinstance(group_by_call, UnderscoreCall):
        raise ChalkParseError("expected a group_by expression")

    group_by_call_parent = group_by_call._chalk__parent
    if not isinstance(group_by_call_parent, UnderscoreAttr):
        raise ChalkParseError("expected a group_by expression")

    group_by_call_attr = group_by_call_parent._chalk__attr
    if group_by_call_attr != "group_by":
        raise ChalkParseError("expected a group_by expression")

    group_key_exprs = group_by_call._chalk__args

    if len(group_by_call._chalk__kwargs) > 0:
        raise ChalkParseError("group_by should not have any keyword arguments, only arguments")

    group_keys: list[str] = []
    for group_key_expr in group_key_exprs:
        group_key = _parse_projection(group_key_expr)
        group_keys.append(group_key)

    # _.transactions.group_by(_.mcc).agg(_.amount.sum())

    has_many_parent = group_by_call_parent._chalk__parent

    # this one has filters in it
    filters: list[UnderscoreFunction] = []
    if isinstance(has_many_parent, UnderscoreItem):
        projections, filters = extract_filters_and_projections(has_many_parent)

        if len(projections) != 0:
            raise ChalkParseError(
                "projections for the group_by should appear in the `.agg` call, like `.agg(_.amount.sum())`"
            )

        has_many_parent = has_many_parent._chalk__parent

    has_many_name = _parse_projection(has_many_parent)

    joined_class, group_key_features, aggregated_feature = _get_has_many_class(
        parent=f.features_cls,
        has_many_feature_name=has_many_name,
        group_names=group_keys,
        aggregated_feature_name=child_feature_name,
    )
    _check_types(
        feature_name=f.name,
        aggregation=aggregation,
        joined_feature=aggregated_feature,
        this_annotation=kind,
    )

    parsed_filters = clean_filters(joined_class, filters)

    if bucket_start.tzinfo is None:
        bucket_start = bucket_start.replace(tzinfo=timezone.utc)
    bucket_start = bucket_start.astimezone(tz=timezone.utc)

    cfg = WindowConfigResolved(
        namespace=joined_class.namespace,
        group_by=group_key_features,
        bucket_duration_seconds=bucket_duration_seconds,
        bucket_start=bucket_start,
        aggregation=aggregation,
        aggregate_on=aggregated_feature,
        pyarrow_dtype=pyarrow_dtype,
        filters=parsed_filters,
        backfill_resolver=_try_parse_resolver_fqn(
            "backfill_resolver",
            f.window_materialization.get("backfill_resolver", None),
        )
        if isinstance(f.window_materialization, dict)
        else None,
        backfill_schedule=f.window_materialization.get("backfill_schedule", None)
        if isinstance(f.window_materialization, dict)
        else None,
        backfill_lookback_duration_seconds=_try_parse_duration(
            "backfill_lookback_duration",
            f.window_materialization.get("backfill_lookback_duration", None),
        )
        if isinstance(f.window_materialization, dict)
        else None,
        backfill_start_time=_try_parse_datetime(
            "backfill_start_time",
            f.window_materialization.get("backfill_start_time", None),
        )
        if isinstance(f.window_materialization, dict)
        else None,
        continuous_resolver=_try_parse_resolver_fqn(
            "continuous_resolver",
            f.window_materialization.get("continuous_resolver", None),
        )
        if isinstance(f.window_materialization, dict)
        else None,
        continuous_buffer_duration_seconds=_try_parse_duration(
            "continuous_buffer_duration",
            f.window_materialization.get("continuous_buffer_duration", None),
        )
        if isinstance(f.window_materialization, dict)
        else None,
    )

    return cfg


def extract_filters_and_projections(
    expr: Underscore,
) -> Tuple[List[UnderscoreAttr], List[UnderscoreFunction]]:
    projections: list[UnderscoreAttr] = []
    filters: list[UnderscoreFunction] = []
    if not isinstance(expr, UnderscoreItem):
        raise ChalkParseError("expected a feature, like `_.amount`, or a filter, like `_.amount > 0`")

    keys = expr._chalk__key
    if not isinstance(keys, tuple):  # pyright: ignore[reportUnnecessaryIsInstance]
        keys = (keys,)

    for k in keys:
        if isinstance(k, UnderscoreFunction):
            filters.append(k)
        elif isinstance(k, UnderscoreAttr):
            projections.append(k)
        else:
            raise ChalkParseError("expected a feature, like `_.amount`, or a filter, like `_.amount > 0`")

    return projections, filters


def clean_filters(joined_class: Type[Features], filters: list[UnderscoreFunction]) -> List[Filter]:
    parsed_filters: List[Filter] = []
    for filt in filters:
        op = filt._chalk__function_name
        left = filt._chalk__args[0]
        right = filt._chalk__args[1]

        if op not in ("==", "!=", ">", "<", ">=", "<=", "in"):
            raise ChalkParseError(f"expected a boolean operation for the filter, like `_.amount > 0`, but found `{op}`")

        if isinstance(left, UnderscoreAttr):
            left_attr = left._chalk__attr
            if left_attr not in ("chalk_window", "chalk_now"):
                try:
                    left = getattr(joined_class, left_attr)
                except Exception:
                    raise ChalkParseError(f"could not find feature '{left_attr}' in the joined class")
            if left_attr == "chalk_now":
                left = Now

        if isinstance(right, UnderscoreAttr):
            right_attr = right._chalk__attr
            if right_attr not in ("chalk_window", "chalk_now"):
                try:
                    right = getattr(joined_class, right_attr)
                except Exception:
                    raise ChalkParseError(f"could not find feature '{right_attr}' in the joined class")
            if right_attr == "chalk_now":
                right = Now

        parsed_filter = Filter(lhs=left, operation=op, rhs=right)
        parsed_filters.append(parsed_filter)

    return parsed_filters


def parse_windowed_materialization(f: Feature) -> WindowConfigResolved | None:
    if f.window_duration is None:
        return None
    aggregation, getitem_expression = _parse_agg_function_call(f.underscore_expression)

    filters: list[UnderscoreFunction] = []
    aggregated_value = None
    if isinstance(getitem_expression, UnderscoreItem):
        agg_on = None

        keys = getitem_expression._chalk__key
        if not isinstance(keys, tuple):  # pyright: ignore[reportUnnecessaryIsInstance]
            keys = (keys,)

        for k in keys:
            if isinstance(k, UnderscoreFunction):
                filters.append(k)
            elif isinstance(k, UnderscoreAttr):
                agg_on = k
            else:
                raise ChalkParseError("expected a feature, like `_.amount`, or a filter, like `_.amount > 0`")

        if isinstance(agg_on, UnderscoreAttr):
            aggregated_value = agg_on._chalk__attr
            if not isinstance(agg_on._chalk__parent, UnderscoreRoot):
                raise ChalkParseError("expected a single feature from the child namespace, like `_.b`")
        elif aggregation != "count":
            raise ChalkParseError(f"missing attribute to aggregate, e.g. `_.a[_.amount].{aggregation}()`")

        child_attr_expression = getitem_expression._chalk__parent
    elif aggregation == "count":
        child_attr_expression = getitem_expression
    else:
        raise ChalkParseError(f"missing attribute to aggregate, e.g. `_.a[_.amount].{aggregation}()`")

    if not isinstance(child_attr_expression, UnderscoreAttr):
        raise ChalkParseError(f"expected reference to has-many feature, like _.children[...].{aggregation}()`")

    child_attr_name = child_attr_expression._chalk__attr
    if not isinstance(child_attr_expression._chalk__parent, UnderscoreRoot):
        raise ChalkParseError(f"expected single feature of the child namespace, like `_.{child_attr_name}`")

    if f.features_cls is None:
        raise ChalkParseError("feature class is None")

    joined_class, group_by_features, aggregated_feature = _get_has_many_class(
        parent=f.features_cls,
        has_many_feature_name=child_attr_name,
        group_names=[],
        aggregated_feature_name=aggregated_value,
    )

    _check_types(
        feature_name=f.window_stem,
        aggregation=aggregation,
        joined_feature=aggregated_feature,
        this_annotation=f.typ.parsed_annotation,
    )

    if f.window_materialization is None:
        raise ChalkParseError("missing 'materialization' in the windowed feature")

    parsed_filters = clean_filters(joined_class, filters)
    bucket_start = datetime.fromtimestamp(0, tz=timezone.utc)
    if f.window_materialization is True:
        bucket_duration = timedelta(seconds=f.window_duration)
    else:
        bucket_duration = f.window_materialization.get("bucket_duration", None)
        assert bucket_duration is None or isinstance(bucket_duration, timedelta)
        found = False
        for bd, values in f.window_materialization.get("bucket_durations", {}).items():
            assert isinstance(bd, timedelta)
            for value in ensure_tuple(values):
                assert isinstance(value, timedelta)
                if int(value.total_seconds()) == f.window_duration:
                    if found and bucket_duration is not None:
                        raise ChalkParseError(
                            (
                                "Multiple bucket durations found for the windowed feature "
                                f"{f.fqn}['{timedelta_to_duration(f.window_duration)}']: "
                                f"{timedelta_to_duration(bd)} and {timedelta_to_duration(bucket_duration)}"
                            )
                        )
                    bucket_duration = bd
                    found = True

        bucket_start = f.window_materialization.get("bucket_start", bucket_start)
        found = False
        for bs, values in f.window_materialization.get("bucket_starts", {}).items():
            assert isinstance(bs, datetime)
            for value in ensure_tuple(values):
                assert isinstance(value, timedelta)
                if int(value.total_seconds()) == f.window_duration:
                    if found:
                        raise ChalkParseError(
                            (
                                "Multiple bucket starts found for the windowed feature "
                                f"{f.fqn}['{timedelta_to_duration(f.window_duration)}']: "
                                f"{bs.isoformat()} and {bucket_start.isoformat()}"
                            )
                        )
                    bucket_start = bs
                    found = True

        if bucket_duration is None:
            raise ChalkParseError(
                f"No bucket duration was found for the window {timedelta_to_duration(f.window_duration)}"
            )

    if bucket_start.tzinfo is None:
        bucket_start = bucket_start.replace(tzinfo=timezone.utc)
    bucket_start = bucket_start.astimezone(tz=timezone.utc)

    return WindowConfigResolved(
        namespace=joined_class.namespace,
        group_by=group_by_features,
        bucket_duration_seconds=int(bucket_duration.total_seconds()),
        bucket_start=bucket_start,
        aggregation=aggregation,
        aggregate_on=aggregated_feature,
        pyarrow_dtype=f.converter.pyarrow_dtype,
        filters=parsed_filters,
        backfill_resolver=_try_parse_resolver_fqn(
            "backfill_resolver",
            f.window_materialization.get("backfill_resolver", None),
        )
        if isinstance(f.window_materialization, dict)
        else None,
        backfill_schedule=f.window_materialization.get("backfill_schedule", None)
        if isinstance(f.window_materialization, dict)
        else None,
        backfill_lookback_duration_seconds=_try_parse_duration(
            "backfill_lookback_duration",
            f.window_materialization.get("backfill_lookback_duration", None),
        )
        if isinstance(f.window_materialization, dict)
        else None,
        backfill_start_time=_try_parse_datetime(
            "backfill_start_time",
            f.window_materialization.get("backfill_start_time", None),
        )
        if isinstance(f.window_materialization, dict)
        else None,
        continuous_resolver=_try_parse_resolver_fqn(
            "continuous_resolver",
            f.window_materialization.get("continuous_resolver", None),
        )
        if isinstance(f.window_materialization, dict)
        else None,
        continuous_buffer_duration_seconds=_try_parse_duration(
            "continuous_buffer_duration",
            f.window_materialization.get("continuous_buffer_duration", None),
        )
        if isinstance(f.window_materialization, dict)
        else None,
    )


def import_all_files(
    file_allowlist: Optional[List[str]] = None,
    project_root: Optional[Path] = None,
    only_sql_files: bool = False,
    check_ignores: bool = True,
    override: bool = False,
) -> List[FailedImport]:
    if project_root is None:
        project_root = get_directory_root()
    if project_root is None:
        return [
            FailedImport(
                filename="",
                module="",
                traceback="Could not find chalk.yaml in this directory or any parent directory",
            )
        ]

    python_files: list[Path] | None = None
    chalk_sql_files: list[str] | None = None

    if file_allowlist is not None:
        python_files = []
        chalk_sql_files = []
        for f in file_allowlist:
            if f.endswith(".py"):
                python_files.append(Path(f))
            elif f.endswith(".chalk.sql"):
                chalk_sql_files.append(f)

    if only_sql_files:
        return import_sql_file_resolvers(project_root, chalk_sql_files, override=override)

    failed_imports: List[FailedImport] = import_all_python_files_from_dir(
        project_root=project_root,
        file_allowlist=python_files,
        check_ignores=check_ignores,
    )
    has_import_errors = len(failed_imports) > 0
    failed_imports.extend(
        import_sql_file_resolvers(
            path=project_root,
            file_allowlist=chalk_sql_files,
            has_import_errors=has_import_errors,
            override=override,
        )
    )

    run_post_import_fixups()

    return failed_imports


def import_sql_file_resolvers(
    path: Path,
    file_allowlist: Optional[List[str]] = None,
    has_import_errors: bool = False,
    override: bool = False,
):
    if file_allowlist is not None:
        sql_resolver_results = get_sql_file_resolvers_from_paths(
            sources=[*BaseSQLSource.registry, *SQLSourceGroup.registry],
            paths=file_allowlist,
            has_import_errors=has_import_errors,
        )
    else:
        sql_resolver_results = get_sql_file_resolvers(
            sql_file_resolve_location=path,
            sources=[*BaseSQLSource.registry, *SQLSourceGroup.registry],
            has_import_errors=has_import_errors,
        )
    failed_imports: List[FailedImport] = []
    for result in sql_resolver_results:
        if result.resolver:
            result.resolver.add_to_registry(override=override)
        if result.errors and not has_import_errors:
            for error in result.errors:
                failed_imports.append(
                    FailedImport(
                        traceback=f"""EXCEPTION in Chalk SQL file resolver '{error.path}':
    {error.display}
""",
                        filename=error.path,
                        module=error.path,
                    )
                )
    return failed_imports


def get_resolver(
    resolver_fqn_or_name: str,
    project_root: Optional[Path] = None,
    only_sql_files: bool = False,
) -> Resolver:
    """
    Returns a resolver by name or fqn, including sql file resolvers.

    Parameters
    ----------
    resolver_fqn_or_name
        A string fqn or name of a resolver. Can also be a filename of sql file resolver
    project_root
        An optional path to import sql file resolvers from.
        If not supplied, will select the root directory of the Chalk project.
    only_sql_files
        If you have already imported all your features, sources, and resolvers, this flag
        can be used to restrict file search to sql file resolvers.

    Returns
    -------
    Resolver
    """
    failed_imports = import_all_files_once(project_root=project_root, only_sql_files=only_sql_files)
    if failed_imports:
        raise ValueError(f"File imports failed: {failed_imports}")
    if resolver_fqn_or_name.endswith(".chalk.sql"):
        resolver_fqn_or_name = resolver_fqn_or_name[: -len(".chalk.sql")]
    maybe_resolver = RESOLVER_REGISTRY.get_resolver(resolver_fqn_or_name)
    if maybe_resolver is not None:
        return maybe_resolver
    raise ValueError(f"No resolver with fqn or name {resolver_fqn_or_name} found")


def _get_py_files_fast(
    resolved_root: Path,
    venv_path: Optional[Path],
    ignore_config: Optional[IgnoreConfig],
) -> Iterable[Path]:
    """
    Gets all the .py files in the resolved_root directory and its subdirectories.
    Faster than the old method we were using because we are skipping the entire
    directory if the directory is determined to be ignored. But if any .gitignore
    or any .chalkignore file has negation, we revert to checking every filepath
    against each .*ignore file.

    :param resolved_root: Project root absolute path
    :param venv_path: Path of the venv folder to skip importing from.
    :param ignore_config: An optional CombinedIgnoreConfig object. If None, we simply don't check for ignores.
    :return: An iterable of Path each representing a .py file
    """

    for dirpath_str, dirnames, filenames in os.walk(resolved_root):
        dirpath = Path(dirpath_str).resolve()

        if (venv_path is not None and venv_path.samefile(dirpath)) or (
            ignore_config
            and not ignore_config.has_negation
            and ignore_config.ignored(os.path.join(str(dirpath), "#"))
            # Hack to make "dir/**" match "/Users/home/dir"
        ):
            dirnames.clear()  # Skip subdirectories
            continue  # Skip files

        for filename in filenames:
            if filename.endswith(".py"):
                filepath = dirpath / filename
                if not ignore_config or not ignore_config.ignored(filepath):
                    yield filepath


CHALK_IMPORT_FLAG: ContextVar[bool] = ContextVar("CHALK_IMPORT_FLAG", default=False)
""" A env var flag to be set to a truthy value during import to catch unsafe operations like ChalkClient().query()
Methods like that should check this env var flag and raise if run inappropriately """


class _UnderscoreValidationError(ValueError):
    """
    This exception class is thrown to indicate that a user-defined underscore expression is incorrect.
    It will not propagate outside of `supplement_diagnostics`.
    """

    ...


class ChalkImporter:
    def __init__(self):
        super().__init__()
        self.errors: Dict[str, FailedImport] = {}
        self.ranges: Dict[str, RangeGQL] = {}
        self.short_tracebacks: Dict[str, str] = {}
        self.repo_files = None

    def add_repo_files(self, repo_files: List[Path]):
        self.repo_files = repo_files

    def add_error(
        self,
        ex_type: Type[BaseException],
        ex_value: BaseException,
        ex_traceback: TracebackType,
        filename: Path,
        module_path: str,
    ):
        tb = traceback.extract_tb(ex_traceback)
        frame = 0
        error_file = str(filename)
        line_number = None
        for i, tb_frame in enumerate(tb):
            tb_filepath = Path(tb_frame.filename).resolve()
            if self.repo_files and tb_filepath in self.repo_files:
                line_number = tb_frame.lineno
                error_file = tb_frame.filename
            if filename == Path(tb_frame.filename).resolve():
                frame = i
        if error_file in self.errors:
            return
        error_message = f"""{(ex_type and ex_type.__name__) or 'Exception'} at '{error_file}{':' + str(line_number) if line_number is not None else ''}'"""
        full_traceback = f"""{error_message}:
{os.linesep.join(traceback.format_tb(ex_traceback)[frame:])}
{ex_type and ex_type.__name__}: {str(ex_value)}
"""
        self.errors[error_file] = FailedImport(
            traceback=full_traceback,
            filename=str(filename),
            module=module_path,
        )
        if line_number is not None:
            line = linecache.getline(str(filename), line_number)
            if line != "":
                self.ranges[error_file] = RangeGQL(
                    start=PositionGQL(
                        line=line_number,
                        character=len(line) - len(line.lstrip()),
                    ),
                    end=PositionGQL(
                        line=line_number,
                        character=max(len(line) - 1, 0),
                    ),
                )
                self.short_tracebacks[error_file] = error_message

    def get_failed_imports(self) -> List[FailedImport]:
        return list(self.errors.values())

    def convert_to_diagnostic(self, failed_import: FailedImport) -> Union[PublishDiagnosticsParams, None]:
        if failed_import.filename == "" or failed_import.filename not in self.ranges:
            return None

        range_ = self.ranges[failed_import.filename]
        traceback_ = self.errors[failed_import.filename].traceback
        builder = DiagnosticBuilder(
            severity=DiagnosticSeverityGQL.Error,
            message=traceback_,
            uri=failed_import.filename,
            range=range_,
            label="failed import",
            code="0",
            code_href=None,
        )
        return PublishDiagnosticsParams(
            uri=failed_import.filename,
            diagnostics=[builder.diagnostic],
        )

    def supplement_diagnostics(
        self,
        failed_imports: List[FailedImport],
        diagnostics: List[PublishDiagnosticsParams],
    ) -> List[PublishDiagnosticsParams]:
        """
        :param failed_imports: Errors encountered when importing customer code. This method converts them into LSP errors.
        :param diagnostics: this list is modified in-place with additioanl diagnostics
        :return: The same object as the input `diagnostics`
        """
        diagnostic_uris = {diagnostic.uri for diagnostic in diagnostics}
        for failed_import in failed_imports:
            if failed_import.filename not in diagnostic_uris:
                diagnostic_or_none = self.convert_to_diagnostic(failed_import)
                if diagnostic_or_none is not None:
                    diagnostics.append(diagnostic_or_none)

        for feature_class in FeatureSetBase.registry.values():
            # Iterate through every class, to find every underscore definition.
            for f in feature_class.features:
                if f.underscore_expression is not None:
                    # Validate that the underscore expression is well-formed.
                    # If it is not well-formed, then an `_UnderscoreValidationError` will
                    # be thrown.
                    try:
                        _supplemental_validate_underscore_expression(
                            state=_SupplementalState(),
                            class_namespace=feature_class,
                            underscore=f.underscore_expression,
                        )
                    except _UnderscoreValidationError as valid_exc:
                        import inspect

                        if f.underscore_expression._chalk_definition_location is not None:
                            # Underscore expressions attempt to obtain a best-effort location.
                            # Currently, the column is always 1, because we do not have a reliable
                            # way to get the column of a caller.
                            underscore_position = PositionGQL(
                                line=f.underscore_expression._chalk_definition_location.line,
                                character=f.underscore_expression._chalk_definition_location.column,
                            )
                            range = RangeGQL(
                                start=underscore_position,
                                end=underscore_position,
                            )
                        else:
                            # This is a fallback used if there is no provided location.
                            range = RangeGQL(
                                start=PositionGQL(line=1, character=1),
                                end=PositionGQL(line=1, character=1),
                            )
                        diagnostics.append(
                            PublishDiagnosticsParams(
                                uri=inspect.getfile(feature_class),
                                diagnostics=[
                                    DiagnosticGQL(
                                        range=range,
                                        message=f"The underscore expression used to define feature '{f.namespace}.{f.name}' is invalid: "
                                        + str(valid_exc),
                                        severity=DiagnosticSeverityGQL.Warning,
                                        code="701",
                                        codeDescription=None,
                                        relatedInformation=None,
                                    )
                                ],
                            )
                        )

        return diagnostics


CHALK_IMPORTER = ChalkImporter()


def import_all_python_files_from_dir(
    project_root: Path,
    check_ignores: bool = True,
    file_allowlist: Optional[List[Path]] = None,
) -> List[FailedImport]:
    project_root = project_root.absolute()

    cwd = os.getcwd()
    os.chdir(project_root)
    # If we don't import both of these, we get in trouble.
    repo_root = Path(project_root)
    resolved_root = repo_root.resolve()
    _logger.debug(f"REPO_ROOT: {resolved_root}")
    sys.path.insert(0, str(resolved_root))
    sys.path.insert(0, str(repo_root.parent.resolve()))
    # Due to the path modifications above, we might have already imported
    # some files under a different module name, and Python doesn't detect
    # duplicate inputs of the same filename under different module names.
    # We can manually detect this by building a set of all absolute
    # filepaths we imported, and then comparing filepaths against this
    # set before attempting to import the module again.
    already_imported_files = {
        Path(v.__file__).resolve(): k
        for (k, v) in sys.modules.copy().items()
        if hasattr(v, "__file__") and isinstance(v.__file__, str)
    }
    token = CHALK_IMPORT_FLAG.set(True)
    try:
        venv = os.environ.get("VIRTUAL_ENV")
        if file_allowlist is not None:
            repo_files = sorted(file_allowlist)
        else:
            venv_path = None if venv is None else Path(venv)
            ignore_config = get_default_combined_ignore_config(resolved_root) if check_ignores else None
            repo_files = sorted(
                list(_get_py_files_fast(resolved_root=resolved_root, venv_path=venv_path, ignore_config=ignore_config))
            )

        CHALK_IMPORTER.add_repo_files(repo_files)
        for filename in repo_files:
            # we want resolved_root in case repo_root contains a symlink
            if filename in already_imported_files:
                _logger.debug(
                    f"Skipping import of '{filename}' since it is already imported as module {already_imported_files[filename]}"
                )
                continue
            module_path = py_path_to_module(filename, resolved_root)
            if module_path.startswith(".eggs") or module_path.startswith("venv") or filename.name == "setup.py":
                continue
            try:
                start = time.perf_counter()
                importlib.import_module(module_path)
                end = time.perf_counter()
                _import_logger.info(f"Imported '{module_path}' in {end-start} seconds")
            except Exception as e:
                if not LSPErrorBuilder.promote_exception(e):
                    ex_type, ex_value, ex_traceback = sys.exc_info()
                    assert ex_type is not None
                    assert ex_value is not None
                    assert ex_traceback is not None
                    CHALK_IMPORTER.add_error(ex_type, ex_value, ex_traceback, filename, module_path)
                    _logger.debug(f"Failed while importing {module_path}", exc_info=True)
            else:
                _logger.debug(f"Imported '{filename}' as module {module_path}")
                already_imported_files[filename] = module_path
    finally:
        CHALK_IMPORT_FLAG.reset(token)
        # Let's remove our added entries in sys.path so we don't pollute it
        sys.path.pop(0)
        sys.path.pop(0)
        # And let's go back to our original directory
        os.chdir(cwd)
    return CHALK_IMPORTER.get_failed_imports()


class _SupplementalState:
    def __init__(self):
        super().__init__()
        self._namespace_features: Dict[str, Dict[str, Feature[Any, Any]]] = {}

    def get_local_features_by_name(
        self,
        class_namespace: type[FeaturesProtocol],
    ) -> Dict[str, Feature[Any, Any]]:
        """
        Returns the features in the provided namespace.
        This function is cached by the namespace's string.
        """

        if class_namespace.namespace in self._namespace_features:
            return self._namespace_features[class_namespace.namespace]

        local_features_by_name: Dict[str, Feature[Any, Any]] = {}
        self._namespace_features[class_namespace.namespace] = local_features_by_name
        for f in class_namespace.features:
            local_features_by_name[f.name] = f

        return local_features_by_name


@dataclass(frozen=True)
class _ScalarExpr:
    dtype: pa.DataType


@dataclass(frozen=True)
class _NamespaceExpr:
    namespace: str


def _supplemental_validate_feature_in_namespace(
    state: _SupplementalState,
    *,
    class_namespace: type[FeaturesProtocol],
    feature_name: str,
    underscore: Underscore,
) -> Union[_ScalarExpr, _NamespaceExpr, None]:
    """
    Resolves references to features.
    """

    # This is a feature reference, in the current namespace.
    class_features = state.get_local_features_by_name(class_namespace)
    if feature_name not in class_features:
        # Complain about this definition!
        raise _UnderscoreValidationError(
            f"'{repr(underscore)}' is invalid because the feature class '{class_namespace.namespace}' does not have a feature named '{underscore._chalk__attr}'"
        )

    feature_definition = class_features[feature_name]
    if feature_definition.is_pseudofeature:
        return None
    if feature_definition.is_windowed:
        return None
    if feature_definition.is_scalar:
        return _ScalarExpr(dtype=feature_definition.converter.pyarrow_dtype)
    if feature_definition.is_has_one:
        foreign_feature_namespace = feature_definition.typ.as_features_cls()
        if foreign_feature_namespace is not None:
            return _NamespaceExpr(namespace=foreign_feature_namespace.namespace)


def _supplemental_validate_underscore_expression(
    state: _SupplementalState,
    *,
    class_namespace: type[FeaturesProtocol],
    underscore: Underscore,
) -> Union[_ScalarExpr, _NamespaceExpr, None]:
    """
    Validates that the provided `underscore` definition is legal when evaluated in `class_namespace`.
    Currently, does not perform type-checking; just checks that all names are in scope.

    If the underscore expression is invalid, raises a `_UnderscoreValidationError` with a message.
    The caller is responsible for converting this exception into a diagnostic message.
    """

    if isinstance(underscore, UnderscoreAttr) and isinstance(underscore._chalk__parent, UnderscoreRoot):
        if underscore._chalk__attr in ("chalk_window", "chalk_now"):
            # This is a special case, which can be skipped in validation.
            return _ScalarExpr(dtype=pa.timestamp("us", tz="UTC"))

        return _supplemental_validate_feature_in_namespace(
            state=state,
            class_namespace=class_namespace,
            feature_name=underscore._chalk__attr,
            underscore=underscore,
        )

    if isinstance(underscore, UnderscoreAttr):
        parent_result = _supplemental_validate_underscore_expression(
            state=state,
            class_namespace=class_namespace,
            underscore=underscore._chalk__parent,
        )

        if isinstance(parent_result, _ScalarExpr):
            if not pa.types.is_struct(parent_result.dtype):
                raise _UnderscoreValidationError(
                    f"the input '{underscore!r}' is a scalar value, so a feature called '{underscore._chalk__attr}' cannot be extracted from it"
                )
            st = cast(pa.StructType, parent_result.dtype)
            field = next((field for field in st if field.name == underscore._chalk__attr), None)
            if field is not None:
                return _ScalarExpr(dtype=field.type)
            raise _UnderscoreValidationError(
                f"could not find the struct field '{underscore._chalk__attr}' among the available fields [{', '.join(f.name for f in st)}]"
            )

        if isinstance(parent_result, _NamespaceExpr):
            if parent_result.namespace not in FeatureSetBase.registry:
                return None

            parent_class_namespace = FeatureSetBase.registry[parent_result.namespace]
            return _supplemental_validate_feature_in_namespace(
                state=state,
                class_namespace=parent_class_namespace,
                feature_name=underscore._chalk__attr,
                underscore=underscore,
            )

        return None

    if isinstance(underscore, UnderscoreFunction):
        if (underscore._chalk__function_name in ("max_by", "min_by") and len(underscore._chalk__args) == 2) or (
            underscore._chalk__function_name in ("max_by_n", "min_by_n") and len(underscore._chalk__args) == 3
        ):
            _supplemental_validate_underscore_expression(
                state=state,
                class_namespace=class_namespace,
                underscore=underscore._chalk__args[0],
            )
            # TODO: second arg is on a different namespace
        else:
            for arg in underscore._chalk__args:
                expr = _supplemental_validate_underscore_expression(
                    state=state,
                    class_namespace=class_namespace,
                    underscore=arg,
                )
                if isinstance(expr, _NamespaceExpr):
                    raise _UnderscoreValidationError(
                        f"the input '{arg!r}' is a feature namespace '{expr.namespace}' which cannot be used as a scalar value"
                    )


def _get_printable_name(typ: Any) -> str:
    try:
        return typ.__name__  # Union types don't have __name__, but have good string representations
    except:
        return str(typ)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/common/v1/offline_query.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.common.v1 import chalk_error_pb2 as chalk_dot_common_dot_v1_dot_chalk__error__pb2
from chalk._gen.chalk.common.v1 import online_query_pb2 as chalk_dot_common_dot_v1_dot_online__query__pb2
from chalk._gen.chalk.expression.v1 import expression_pb2 as chalk_dot_expression_dot_v1_dot_expression__pb2
from chalk._gen.chalk.graph.v1 import graph_pb2 as chalk_dot_graph_dot_v1_dot_graph__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n#chalk/common/v1/offline_query.proto\x12\x0f\x63halk.common.v1\x1a!chalk/common/v1/chalk_error.proto\x1a"chalk/common/v1/online_query.proto\x1a$chalk/expression/v1/expression.proto\x1a\x1a\x63halk/graph/v1/graph.proto\x1a\x1cgoogle/protobuf/struct.proto"\xdc\x01\n\x1dOfflineQueryRecomputeFeatures\x12 \n\x0b\x61ll_or_none\x18\x01 \x01(\x08H\x00R\tallOrNone\x12_\n\x0c\x66\x65\x61ture_list\x18\x02 \x01(\x0b\x32:.chalk.common.v1.OfflineQueryRecomputeFeatures.FeatureListH\x00R\x0b\x66\x65\x61tureList\x1a\x30\n\x0b\x46\x65\x61tureList\x12!\n\x0c\x66\x65\x61ture_list\x18\x01 \x03(\tR\x0b\x66\x65\x61tureListB\x06\n\x04impl"\x80\x01\n\x13OfflineQueryExplain\x12\x18\n\x06truthy\x18\x01 \x01(\x08H\x00R\x06truthy\x12?\n\x04only\x18\x02 \x01(\x0b\x32).chalk.common.v1.OfflineQueryExplain.OnlyH\x00R\x04only\x1a\x06\n\x04OnlyB\x06\n\x04impl"a\n\x11OfflineQueryInput\x12\x18\n\x07\x63olumns\x18\x01 \x03(\tR\x07\x63olumns\x12\x32\n\x06values\x18\x02 \x03(\x0b\x32\x1a.google.protobuf.ListValueR\x06values"V\n\x18OfflineQueryInputSharded\x12:\n\x06inputs\x18\x01 \x03(\x0b\x32".chalk.common.v1.OfflineQueryInputR\x06inputs"\xbd\x02\n\x12OfflineQueryInputs\x12\'\n\x0e\x66\x65\x61ther_inputs\x18\x01 \x01(\x0cH\x00R\rfeatherInputs\x12K\n\tno_inputs\x18\x02 \x01(\x0b\x32,.chalk.common.v1.OfflineQueryInputs.NoInputsH\x00R\x08noInputs\x12I\n\rsingle_inputs\x18\x03 \x01(\x0b\x32".chalk.common.v1.OfflineQueryInputH\x00R\x0csingleInputs\x12R\n\x0esharded_inputs\x18\x04 \x01(\x0b\x32).chalk.common.v1.OfflineQueryInputShardedH\x00R\rshardedInputs\x1a\n\n\x08NoInputsB\x06\n\x04impl"\xf8\x0e\n\x13OfflineQueryRequest\x12;\n\x06inputs\x18\x01 \x01(\x0b\x32#.chalk.common.v1.OfflineQueryInputsR\x06inputs\x12\x18\n\x07outputs\x18\x02 \x03(\tR\x07outputs\x12)\n\x10required_outputs\x18\x03 \x03(\tR\x0frequiredOutputs\x12-\n\x12\x64\x65stination_format\x18\x04 \x01(\tR\x11\x64\x65stinationFormat\x12\x1b\n\x06\x62ranch\x18\x05 \x01(\tH\x00R\x06\x62ranch\x88\x01\x01\x12&\n\x0c\x64\x61taset_name\x18\x06 \x01(\tH\x01R\x0b\x64\x61tasetName\x88\x01\x01\x12]\n\x12recompute_features\x18\x07 \x01(\x0b\x32..chalk.common.v1.OfflineQueryRecomputeFeaturesR\x11recomputeFeatures\x12*\n\x11store_plan_stages\x18\x08 \x01(\x08R\x0fstorePlanStages\x12>\n\x07\x66ilters\x18\x0b \x03(\x0b\x32$.chalk.expression.v1.LogicalExprNodeR\x07\x66ilters\x12+\n\x0fspine_sql_query\x18\x0c \x01(\tH\x02R\rspineSqlQuery\x88\x01\x01\x12$\n\x0bmax_samples\x18\x65 \x01(\x05H\x03R\nmaxSamples\x88\x01\x01\x12\x30\n\x12max_cache_age_secs\x18\x66 \x01(\x05H\x04R\x0fmaxCacheAgeSecs\x88\x01\x01\x12\x42\n\x07\x65xplain\x18g \x01(\x0b\x32$.chalk.common.v1.OfflineQueryExplainB\x02\x18\x01R\x07\x65xplain\x12;\n\x08\x65xplain2\x18j \x01(\x0b\x32\x1f.chalk.common.v1.ExplainOptionsR\x08\x65xplain2\x12\x12\n\x04tags\x18h \x03(\tR\x04tags\x12*\n\x0e\x63orrelation_id\x18i \x01(\tH\x05R\rcorrelationId\x88\x01\x01\x12\x34\n\x16required_resolver_tags\x18k \x03(\tR\x14requiredResolverTags\x12;\n\x17observed_at_lower_bound\x18\xc9\x01 \x01(\tH\x06R\x14observedAtLowerBound\x88\x01\x01\x12;\n\x17observed_at_upper_bound\x18\xca\x01 \x01(\tH\x07R\x14observedAtUpperBound\x88\x01\x01\x12\x62\n\x0fplanner_options\x18\xcb\x01 \x03(\x0b\x32\x38.chalk.common.v1.OfflineQueryRequest.PlannerOptionsEntryR\x0eplannerOptions\x12\'\n\x0cstore_online\x18\xcc\x01 \x01(\x08H\x08R\x0bstoreOnline\x88\x01\x01\x12)\n\rstore_offline\x18\xcd\x01 \x01(\x08H\tR\x0cstoreOffline\x88\x01\x01\x12:\n\x16use_multiple_computers\x18\xce\x01 \x01(\x08H\nR\x14useMultipleComputers\x88\x01\x01\x12#\n\nnum_shards\x18\xcf\x01 \x01(\x05H\x0bR\tnumShards\x88\x01\x01\x12%\n\x0bnum_workers\x18\xd0\x01 \x01(\x05H\x0cR\nnumWorkers\x88\x01\x01\x12\\\n\rquery_context\x18\xd1\x01 \x03(\x0b\x32\x36.chalk.common.v1.OfflineQueryRequest.QueryContextEntryR\x0cqueryContext\x12G\n\roverlay_graph\x18\xd2\x01 \x01(\x0b\x32\x1c.chalk.graph.v1.OverlayGraphH\rR\x0coverlayGraph\x88\x01\x01\x1aY\n\x13PlannerOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value:\x02\x38\x01\x1aW\n\x11QueryContextEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value:\x02\x38\x01\x42\t\n\x07_branchB\x0f\n\r_dataset_nameB\x12\n\x10_spine_sql_queryB\x0e\n\x0c_max_samplesB\x15\n\x13_max_cache_age_secsB\x11\n\x0f_correlation_idB\x1a\n\x18_observed_at_lower_boundB\x1a\n\x18_observed_at_upper_boundB\x0f\n\r_store_onlineB\x10\n\x0e_store_offlineB\x19\n\x17_use_multiple_computersB\r\n\x0b_num_shardsB\x0e\n\x0c_num_workersB\x10\n\x0e_overlay_graph"\xce\x01\n\x12\x43olumnMetadataList\x12N\n\x08metadata\x18\x01 \x03(\x0b\x32\x32.chalk.common.v1.ColumnMetadataList.ColumnMetadataR\x08metadata\x1ah\n\x0e\x43olumnMetadata\x12\x1f\n\x0b\x66\x65\x61ture_fqn\x18\x01 \x01(\tR\nfeatureFqn\x12\x1f\n\x0b\x63olumn_name\x18\x02 \x01(\tR\ncolumnName\x12\x14\n\x05\x64type\x18\x03 \x01(\tR\x05\x64type"\xf0\x01\n\x1aGetOfflineQueryJobResponse\x12\x1f\n\x0bis_finished\x18\x01 \x01(\x08R\nisFinished\x12\x18\n\x07version\x18\x02 \x01(\x05R\x07version\x12\x12\n\x04urls\x18\x03 \x03(\tR\x04urls\x12\x33\n\x06\x65rrors\x18\x04 \x03(\x0b\x32\x1b.chalk.common.v1.ChalkErrorR\x06\x65rrors\x12\x42\n\x07\x63olumns\x18\x05 \x01(\x0b\x32#.chalk.common.v1.ColumnMetadataListH\x00R\x07\x63olumns\x88\x01\x01\x42\n\n\x08_columnsB\x86\x01\n\x13\x63om.chalk.common.v1B\x11OfflineQueryProtoP\x01\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43halk.Common.V1\xca\x02\x0f\x43halk\\Common\\V1\xe2\x02\x1b\x43halk\\Common\\V1\\GPBMetadata\xea\x02\x11\x43halk::Common::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.common.v1.offline_query_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.common.v1B\021OfflineQueryProtoP\001\242\002\003CCX\252\002\017Chalk.Common.V1\312\002\017Chalk\\Common\\V1\342\002\033Chalk\\Common\\V1\\GPBMetadata\352\002\021Chalk::Common::V1"
    _globals["_OFFLINEQUERYREQUEST_PLANNEROPTIONSENTRY"]._options = None
    _globals["_OFFLINEQUERYREQUEST_PLANNEROPTIONSENTRY"]._serialized_options = b"8\001"
    _globals["_OFFLINEQUERYREQUEST_QUERYCONTEXTENTRY"]._options = None
    _globals["_OFFLINEQUERYREQUEST_QUERYCONTEXTENTRY"]._serialized_options = b"8\001"
    _globals["_OFFLINEQUERYREQUEST"].fields_by_name["explain"]._options = None
    _globals["_OFFLINEQUERYREQUEST"].fields_by_name["explain"]._serialized_options = b"\030\001"
    _globals["_OFFLINEQUERYRECOMPUTEFEATURES"]._serialized_start = 224
    _globals["_OFFLINEQUERYRECOMPUTEFEATURES"]._serialized_end = 444
    _globals["_OFFLINEQUERYRECOMPUTEFEATURES_FEATURELIST"]._serialized_start = 388
    _globals["_OFFLINEQUERYRECOMPUTEFEATURES_FEATURELIST"]._serialized_end = 436
    _globals["_OFFLINEQUERYEXPLAIN"]._serialized_start = 447
    _globals["_OFFLINEQUERYEXPLAIN"]._serialized_end = 575
    _globals["_OFFLINEQUERYEXPLAIN_ONLY"]._serialized_start = 561
    _globals["_OFFLINEQUERYEXPLAIN_ONLY"]._serialized_end = 567
    _globals["_OFFLINEQUERYINPUT"]._serialized_start = 577
    _globals["_OFFLINEQUERYINPUT"]._serialized_end = 674
    _globals["_OFFLINEQUERYINPUTSHARDED"]._serialized_start = 676
    _globals["_OFFLINEQUERYINPUTSHARDED"]._serialized_end = 762
    _globals["_OFFLINEQUERYINPUTS"]._serialized_start = 765
    _globals["_OFFLINEQUERYINPUTS"]._serialized_end = 1082
    _globals["_OFFLINEQUERYINPUTS_NOINPUTS"]._serialized_start = 1064
    _globals["_OFFLINEQUERYINPUTS_NOINPUTS"]._serialized_end = 1074
    _globals["_OFFLINEQUERYREQUEST"]._serialized_start = 1085
    _globals["_OFFLINEQUERYREQUEST"]._serialized_end = 2997
    _globals["_OFFLINEQUERYREQUEST_PLANNEROPTIONSENTRY"]._serialized_start = 2546
    _globals["_OFFLINEQUERYREQUEST_PLANNEROPTIONSENTRY"]._serialized_end = 2635
    _globals["_OFFLINEQUERYREQUEST_QUERYCONTEXTENTRY"]._serialized_start = 2637
    _globals["_OFFLINEQUERYREQUEST_QUERYCONTEXTENTRY"]._serialized_end = 2724
    _globals["_COLUMNMETADATALIST"]._serialized_start = 3000
    _globals["_COLUMNMETADATALIST"]._serialized_end = 3206
    _globals["_COLUMNMETADATALIST_COLUMNMETADATA"]._serialized_start = 3102
    _globals["_COLUMNMETADATALIST_COLUMNMETADATA"]._serialized_end = 3206
    _globals["_GETOFFLINEQUERYJOBRESPONSE"]._serialized_start = 3209
    _globals["_GETOFFLINEQUERYJOBRESPONSE"]._serialized_end = 3449
# @@protoc_insertion_point(module_scope)
